from tests.unit.mocks import GlobusTransferTaskResponse
from pilot.transfer_log import TransferLog


def test_add_transfer_log(mock_config):
    tl = TransferLog(mock_config)
    cfg = tl.config.load()
    assert cfg['transfer_log'] == {}
    assert tl.get_log() == []
    gccr = GlobusTransferTaskResponse()
    tl.add_log(gccr, 'foo/bar')
    assert cfg['transfer_log'] != {}
    assert list(cfg['transfer_log'].keys()) == ['0']
    assert cfg['transfer_log']['0'] is not None


def test_get_transfer_log(mock_config):
    tl = TransferLog(mock_config)
    gccr = GlobusTransferTaskResponse()
    tl.add_log(gccr, 'foo/bar')

    tlog = tl.get_log()
    assert len(tlog) == 1
    mylog = tlog[0]
    assert set(mylog.keys()) == {'dataframe', 'id', 'task_id',
                                 'start_time', 'status'}


def test_update_transfer_log(mock_config):
    tl = TransferLog(mock_config)
    gccr = GlobusTransferTaskResponse()
    tl.add_log(gccr, 'foo/bar')
    tl.update_log(gccr.data['task_id'], 'complete')
    tlog = tl.get_log_by_task(gccr.data['task_id'])
    assert tlog['status'] == 'complete'



import os
import sys
import json
import logging
import click
import globus_sdk
import datetime
import pilot
import traceback
import contextlib
import pathlib
from pilot.exc import HTTPSClientException, InvalidField, ExitCodes
from pilot.search_parse import get_size
from pilot.commands.endpoint_utils import test_local_endpoint
from jsonschema.exceptions import ValidationError

log = logging.getLogger(__name__)

# Warn people things will take a while when filesize exceeds 1GB
BIG_SIZE_WARNING = 2 ** 30


def load_json(filename):
    if filename is not None and os.path.exists(filename):
        with open(filename) as fh:
            return json.load(fh)
    return {}


@click.command(help='Upload dataframe to location on Globus and categorize it '
                    'in search')
@click.argument('dataframe',
                type=click.Path(exists=True, file_okay=True, dir_okay=True,
                                readable=True, resolve_path=True))
@click.argument('destination', type=click.Path(), required=False)
@click.option('-j', '--json', 'metadata', type=click.Path(),
              help='Add custom metadata. For field reference, go to:\n'
                   'https://github.com/globusonline/pilot1-tools '
                   'and navigate to doc under `docs/reference.rst`')
@click.option('-u', '--update/--no-update', default=False,
              help='Overwrite an existing dataframe and increment the version')
@click.option('--gcp/--no-gcp', default=True,
              help='Use Globus Connect Personal to start a transfer instead '
                   'of uploading using direct HTTP')
@click.option('--dry-run', is_flag=True, default=False,
              help='Do checks and validation but do not upload/ingest. ')
@click.option('--verbose', is_flag=True, default=False)
@click.option('--no-analyze', is_flag=True, default=False,
              help='Analyze the field to collect additional metadata.')
@click.option('--foreign-keys', 'foreign_keys', type=click.Path(),
              help='File containing links to other search records.')
def upload(dataframe, destination, metadata, gcp, update, dry_run,
           verbose, no_analyze, foreign_keys):
    """
    Create a search entry and upload this file to the GCS Endpoint.
    """
    with pilot_code_handler(dataframe, destination, verbose):
        pc = pilot.commands.get_pilot_client()
        transport = 'globus' if gcp else 'http'
        basename = os.path.basename(dataframe)
        click.secho('Uploading {} using {}... '.format(basename,
                                                       transport))
        test_local_endpoint()
        stats = pc.upload(dataframe, destination, metadata=load_json(metadata),
                          globus=gcp, update=update, dry_run=dry_run,
                          skip_analysis=no_analyze,
                          foreign_keys=load_json(foreign_keys))
        short_path = os.path.join(destination, basename)
        if dry_run:
            raise pilot.exc.DryRun(stats=stats, verbose=verbose)
        elif not stats['metadata_modified']:
            raise pilot.exc.NoChangesNeeded(fmt=[short_path])
        click.secho('A transfer has been queued, see `pilot status` for '
                    'an update of the transfer.', fg='green')
        url = pc.get_portal_url(short_path)
        if url:
            click.echo('You can view your new record here: \n{}'.format(url))


@click.command(help='Register an existing dataframe in search', hidden=True)
@click.argument('dataframe',
                type=click.Path(exists=True, file_okay=True, dir_okay=True,
                                readable=True, resolve_path=True),)
@click.argument('destination', type=click.Path(), required=False)
@click.option('-j', '--json', 'metadata', type=click.Path(),
              help='Metadata in JSON format')
@click.option('-u', '--update/--no-update', default=False,
              help='Overwrite an existing dataframe and increment the version')
@click.option('--dry-run', is_flag=True, default=False,
              help='Do checks and validation but do not upload/ingest. ')
@click.option('--verbose', is_flag=True, default=False)
@click.option('--no-analyze', is_flag=True, default=False,
              help='Analyze the field to collect additional metadata.')
@click.option('--foreign-keys', 'foreign_keys', type=click.Path(),
              help='File containing links to other search records.')
def register(dataframe, destination, metadata, update, dry_run, verbose,
             no_analyze, foreign_keys):
    """
    Create a search entry for a pre-existing file
    """
    with pilot_code_handler(dataframe, destination, verbose):
        pc = pilot.commands.get_pilot_client()
        click.secho('Registering {}... '.format(dataframe))
        short_path = os.path.join(destination, os.path.basename(dataframe))
        stats = pc.register(dataframe, destination,
                            metadata=load_json(metadata), update=update,
                            dry_run=dry_run, skip_analysis=no_analyze,
                            foreign_keys=load_json(foreign_keys))
        if dry_run:
            raise pilot.exc.DryRun(stats=stats, verbose=verbose)
        elif not stats['metadata_modified']:
            raise pilot.exc.NoChangesNeeded(fmt=[short_path])
        click.secho('Success!', fg='green')
        url = pc.get_portal_url(short_path)
        if url:
            click.echo('You can view your new record here: \n{}'.format(url))


@contextlib.contextmanager
def pilot_code_handler(dataframe, destination, verbose):
    """
    Handle exceptions for a `pc.register` or `pc.upload` call. If exceptions
    happen, this handler will cleanly catch the exception and print click
    output for the result. Only code exceptions are caught, all other
    exceptions fall through.
    ** parameters **
      ``short_name`` (*string*) the short-hard representation of the relative
      path on the remote endpoint
      ``verbose`` (*bool*) print verbose output
    """
    try:
        pc = pilot.commands.get_pilot_client()
        if not destination:
            dirs = [n for n, d in pc.ls('', extended=True).items()
                    if d['type'] == 'dir']
            raise pilot.exc.NoDestinationProvided(fmt=[dirs])
        yield
    except (ValidationError, InvalidField) as e:
        log.exception(e)
        click.secho('Error Validating Metadata: {}'.format(e), fg='red')
        sys.exit(ExitCodes.INVALID_METADATA)
    except pilot.exc.RecordExists as re:
        last_updated = re.previous_metadata['dc']['dates'][-1]['date']
        dt = datetime.datetime.strptime(last_updated, '%Y-%m-%dT%H:%M:%S.%fZ')
        short_path = os.path.join(destination, os.path.basename(dataframe))
        click.echo('Existing record found for {}, specify -u to update.\n'
                   'Last updated: {: %A, %b %d, %Y}'
                   ''.format(short_path, dt))
        sys.exit(re.CODE)
    except pilot.exc.DryRun as dr:
        pc = pilot.commands.get_pilot_client()
        short_path = os.path.join(destination, os.path.basename(dataframe))
        stats = dr.stats
        stats.update({'subject': pc.get_subject_url(short_path),
                      'url': pc.get_globus_http_url(short_path)})
        click.secho('Success! (Dry Run -- No changes made.)', fg='green')
        click.echo('Pre-existing record: {record_exists}\n'
                   'Files Modified: {files_modified}\n'
                   'Metadata Modified: {metadata_modified}\n'
                   'Version: {version} (Would be updated to {new_version})\n'
                   'Search Subject: {subject}\n'
                   'URL: {url}\n'
                   .format(**stats)
                   )
        if verbose:
            _, metadata = dr.verbose_output
            click.echo(json.dumps(metadata, indent=2))
    except pilot.exc.PilotCodeException as pce:
        if verbose:
            click.echo(pce.verbose_output)
        fg = 'green' if pce.CODE == ExitCodes.SUCCESS else 'yellow'
        click.secho(str(pce), err=True, fg=fg)
        sys.exit(pce.CODE)
    except pilot.exc.AnalysisException as ae:
        click.secho('Error analyzing {}, skipping...'.format(dataframe),
                    fg='yellow')
        if verbose:
            traceback.print_exception(*ae.original_exc_info)
        else:
            click.secho('(Use --verbose to see full error)', fg='yellow')
    except globus_sdk.exc.TransferAPIError as tapie:
        click.secho(tapie.message, fg='yellow')


@click.command(help='Download a file to your local directory.')
@click.argument('path', type=click.Path())
@click.option('--overwrite/--no-overwrite', default=True)
@click.option('--range', help='Download only part of a file. '
                              'Ex: bytes=0-1, 4-5')
def download(path, overwrite, range):
    # TODO -- Downloading single files within mfes is BROKEN! This instead
    # casues all files to be downloaded
    pc = pilot.commands.get_pilot_client()
    fname = os.path.basename(path)
    if os.path.exists(fname) and not overwrite:
        click.echo('Aborted! File {} would be overwritten.'.format(fname))
        return
    try:
        ent = pc.get_search_entry(path, resolve_collections=True,
                                  precise=False)
        if not ent:
            click.secho('No record exists for {}.'.format(path), fg='yellow')
            sys.exit(pilot.exc.ExitCodes.NO_RECORD_EXISTS)
        http_path = pc.get_globus_http_url(path)
        to_download = [f for f in ent['files'] if http_path in f['url']]
        if len(to_download) > 1:
            click.secho('Downloading {} files, totalling {}'.format(
                len(to_download), get_size({'files': to_download})))
        for file_ent in to_download:
            base_path = os.path.dirname(http_path)
            dest = file_ent['url'].replace(base_path, '').lstrip('/')
            if file_ent.get('length') == 0:
                click.secho('{} is an empty file'.format(dest))
                pathlib.Path(dest).touch()
                continue
            r_content = pc.download_parts(file_ent['url'], dest=dest,
                                          project=None, range=range)
            params = {'label': 'Downloading {}'.format(dest),
                      'length': file_ent['length'], 'show_pos': True}
            with click.progressbar(**params) as bar:
                for bytes_written in r_content:
                    bar.update(bytes_written)
    except HTTPSClientException as hce:
        log.exception(hce)
        if hce.http_status == 404:
            click.secho(f'File not found {path}', fg='red')
        else:
            click.secho('An unexpected error occurred, please contact your '
                        'system administrator', fg='red')


@click.command(help='The new path to create')
@click.argument('path', type=click.Path())
def mkdir(path):
    pc = pilot.commands.get_pilot_client()
    if not pc.is_logged_in():
        click.echo('You are not logged in.')
        return
    try:
        pc.mkdir(path)
        click.secho('Created directory {}'.format(path), fg='green')
    except globus_sdk.exc.TransferAPIError as tapie:
        if tapie.code == 'ExternalError.MkdirFailed.Exists':
            click.secho('Directory already exists')
        elif 'No such file or directory' in tapie.message:
            click.secho('Parent directory does not exist', fg='red')
        elif tapie.code == 'EndpointPermissionDenied':
            click.secho('Permission Denied, you do not have access to create '
                        'directories on {}'.format(pc.project.current),
                        fg='red')
        else:
            log.exception(tapie)
            click.secho('Unknown error, please send this to your system '
                        'administrator', fg='red')



import os
import pytest
import json
from unittest.mock import Mock

import globus_sdk
import jsonschema
from pilot.search import scrape_metadata
from pilot import exc, analysis
from pilot.exc import AnalysisException
from tests.unit.mocks import COMMANDS_FILE_BASE_DIR, MULTI_FILE_DIR

EMPTY_TEST_FILE = os.path.join(COMMANDS_FILE_BASE_DIR,
                               'test_file_zero_length.txt')
EMTPY_TEST_FILE_META = os.path.join(COMMANDS_FILE_BASE_DIR,
                                    'empty_file_metadata.json')
SMALL_TEST_FILE = os.path.join(COMMANDS_FILE_BASE_DIR, 'test_file_small.txt')
CUSTOM_METADATA = os.path.join(COMMANDS_FILE_BASE_DIR, 'custom_metadata.json')
INVALID_METADATA = os.path.join(COMMANDS_FILE_BASE_DIR,
                                'invalid_metadata.json')


def test_file_upload(mock_cli, mock_transfer_log, mock_search_client):
    # Should not raise errors
    metadata = mock_cli.upload(EMPTY_TEST_FILE, 'my_folder')
    assert set(metadata['new_metadata']) == {'files', 'dc', 'project_metadata'}
    assert metadata['files_modified'] is True
    assert metadata['metadata_modified'] is True
    assert metadata['protocol'] == 'globus'
    from pprint import pprint
    pprint(metadata)
    assert mock_search_client.ingest.called
    # This is a nice way to ensure the transfer was initiated
    assert mock_transfer_log.called


def test_dir_upload(mock_cli, mock_transfer_log):
    # Should not raise errors
    metadata = mock_cli.upload(MULTI_FILE_DIR, 'my_folder')['new_metadata']
    assert set(metadata) == {'files', 'dc', 'project_metadata'}
    assert len(metadata['files']) == 4
    expected_paths = [
        'my_folder/multi_file/text_metadata.txt',
        'my_folder/multi_file/folder/tsv1.tsv',
        'my_folder/multi_file/folder/tinyimage.png',
        'my_folder/multi_file/folder/folder2/tsv2.tsv',
    ]
    expected_urls = [mock_cli.get_globus_http_url(u) for u in expected_paths]
    urls = [f['url'] for f in metadata['files']]
    for url in urls:
        assert url in expected_urls


def test_update_mfe_with_file(mock_cli, mock_transfer_log,
                              mock_multi_file_result):
    sub = mock_cli.get_subject_url('my_folder/multi_file')
    mock_multi_file_result['gmeta'][0]['subject'] = sub
    gse = Mock(return_value=mock_multi_file_result['gmeta'])
    mock_cli.list_entries = gse
    metadata = mock_cli.upload(EMPTY_TEST_FILE, 'my_folder/multi_file',
                               update=True)['new_metadata']
    assert len(mock_multi_file_result['gmeta'][0]['content'][0]['files']) == 4
    assert len(metadata['files']) == 5
    urls = [f['url'] for f in metadata['files']]
    new_url = mock_cli.get_globus_http_url('my_folder/multi_file/' +
                                           os.path.basename(EMPTY_TEST_FILE))
    assert new_url in urls


def test_update_mfe_with_dir(mock_cli, mock_transfer_log,
                             mock_multi_file_result):
    sub = mock_cli.get_subject_url('my_folder/multi_file')
    mock_multi_file_result['gmeta'][0]['subject'] = sub
    gse = Mock(return_value=mock_multi_file_result['gmeta'])
    mock_cli.list_entries = gse
    metadata = mock_cli.upload(MULTI_FILE_DIR, 'my_folder/multi_file',
                               update=True)['new_metadata']
    assert len(mock_multi_file_result['gmeta'][0]['content'][0]['files']) == 4
    assert len(metadata['files']) == 8


def test_upload_in_dir_with_similar_record(mock_cli, mock_search_result):
    """
    This was to fix a bug where uploading similar records would cause conflicts

    * my_folder/simple_tsv
    * my_folder/empty_test_file.txt
    """
    sub = mock_cli.get_subject_url('my_folder/simple_tsv')
    mock_search_result['subject'] = sub
    gse = Mock(return_value=[mock_search_result])
    mock_cli.list_entries = gse
    # should not raise RecordExists Exception
    mock_cli.upload(EMPTY_TEST_FILE, 'my_folder')


def test_upload_without_destination(mock_cli):
    with pytest.raises(exc.NoDestinationProvided):
        mock_cli.upload(EMPTY_TEST_FILE, None)


def test_upload_to_nonexistant_dir(mock_cli, mock_transfer_error):
    mock_transfer_error.code = 'ClientError.NotFound'
    mock_cli.ls = Mock(side_effect=globus_sdk.exc.TransferAPIError)
    with pytest.raises(exc.DirectoryDoesNotExist):
        mock_cli.upload(EMPTY_TEST_FILE, 'my_folder')


def test_upload_destination_is_record(mock_cli, mock_multi_file_result):
    mock_cli.search.return_value = mock_multi_file_result
    with pytest.raises(exc.RecordExists):
        mock_cli.upload(EMPTY_TEST_FILE, '/multi_file/foo')


def test_upload_unexpected_ls_error(mock_cli, mock_transfer_error):
    mock_transfer_error.code = 'UnexpectedError'
    mock_cli.ls = Mock(side_effect=globus_sdk.exc.TransferAPIError)
    with pytest.raises(exc.GlobusTransferError):
        mock_cli.upload(EMPTY_TEST_FILE, 'my_folder')


def test_upload_with_custom_metadata(mock_cli):
    cust_meta = {'custom_key': 'custom_value'}
    stats = mock_cli.upload(EMPTY_TEST_FILE, 'my_folder', metadata=cust_meta)
    meta = stats['new_metadata']
    assert 'custom_key' in meta['project_metadata']
    assert meta['project_metadata']['custom_key'] == 'custom_value'


def test_upload_analyze_error(mock_cli, monkeypatch):
    mock_exc = Mock(side_effect=AnalysisException('fail!', None))
    monkeypatch.setattr(analysis, 'analyze_dataframe', mock_exc)
    with pytest.raises(exc.AnalysisException):
        mock_cli.upload(EMPTY_TEST_FILE, 'my_folder')


def test_upload_validation_error(mock_cli, mock_transfer_log):
    invalid_m = {'formats': [1234]}
    with pytest.raises(jsonschema.exceptions.ValidationError):
        mock_cli.upload(EMPTY_TEST_FILE, 'my_folder', metadata=invalid_m)


def test_no_update_needed(mock_cli, mock_transfer_log, mock_search_client):
    basen = os.path.basename(EMPTY_TEST_FILE)
    url = mock_cli.get_globus_http_url(basen)
    meta = scrape_metadata(EMPTY_TEST_FILE, url, mock_cli.profile,
                           'foo-project')
    entry = {'content': [meta], 'subject': mock_cli.get_subject_url(basen)}
    mock_cli.list_entries = Mock(return_value=[entry])
    mock_cli.upload(EMPTY_TEST_FILE, '/', update=True)
    assert not mock_search_client.ingest.called
    assert not mock_transfer_log.called


def test_upload_record_exists(mock_cli):
    url = mock_cli.get_globus_http_url('my_folder/test_file_zero_length.txt')
    sub = mock_cli.get_subject_url('my_folder')
    meta = scrape_metadata(EMPTY_TEST_FILE, url, mock_cli.profile, 'foo')
    entry = {'content': [meta], 'subject': sub}
    mock_cli.list_entries = Mock(return_value=[entry])
    with pytest.raises(exc.RecordExists):
        mock_cli.upload(SMALL_TEST_FILE, 'my_folder')


def test_upload_dry_run(mock_cli):
    stats = mock_cli.upload(SMALL_TEST_FILE, 'my_folder', dry_run=True)
    assert not stats['ingest']
    assert not stats['upload']
    assert stats['files_modified'] is True
    assert stats['metadata_modified'] is True
    assert stats['protocol'] == 'globus'
    assert stats['version'] is None
    assert stats['record_exists'] is False
    assert stats['new_version'] == '1'


def test_dataframe_up_to_date(mock_cli, mock_transfer_log, monkeypatch):
    """Update metadata but not the actual file"""
    sub = mock_cli.get_subject_url(os.path.basename(EMPTY_TEST_FILE))
    with open(EMTPY_TEST_FILE_META) as f:
        le = Mock(return_value=[
            {'content': [json.load(f)],
             'subject': sub}
        ])
        mock_cli.list_entries = le
    new_meta = {"custom_metadata_key": "custom_metadata_value"}
    res = mock_cli.upload(EMPTY_TEST_FILE, '/', metadata=new_meta, update=True)
    assert res['record_exists']
    assert res['metadata_modified'] is True
    assert res['files_modified'] is False
    assert not globus_sdk.TransferData.called


def test_upload_local_endpoint_not_set(mock_cli, mock_profile):

    mock_cli.profile.save_option('local_endpoint', None, section='profile')
    with pytest.raises(exc.NoLocalEndpointSet):
        mock_cli.upload(SMALL_TEST_FILE, 'my_folder', globus=True)


def test_upload_gcp_log(mock_cli, mock_transfer_log):
    mock_cli.get_full_search_entry.return_value = {}
    mock_cli.get_transfer_client().submit_transfer.return_value = {}
    mock_cli.upload(SMALL_TEST_FILE, 'my_folder', globus=True)
    assert mock_transfer_log.called



import logging
import uuid
import sys
import click

from fair_research_login import ScopesMismatch

from pilot import commands, exc
from pilot.commands.project.project import project_command
from pilot.context import DEFAULT_PROJECTS_CACHE_TIMEOUT
from pilot.commands import input_validation

log = logging.getLogger(__name__)


@click.group(name='index', help='Set or display index information',
             invoke_without_command=True)
@click.pass_context
def index_command(ctx):
    pc = commands.get_pilot_client()

    if ctx.invoked_subcommand is None:
        click.echo('Set index with "pilot index set <index_uuid>|'
                   '<index_name>"')
        contexts = pc.context.load_all()
        fmt = '{} {}'
        for context in contexts:
            if context == pc.context.current:
                click.secho(fmt.format('*', context), fg='green')
            else:
                click.echo(fmt.format(' ', context))


@index_command.command(help='Set Pilot to use a different search index',
                       name='set')
@click.argument('index_name')
@click.pass_context
def set_index(ctx, index_name):
    pc = commands.get_pilot_client()
    if not pc.context.get_context(index_name):
        log.debug(f'No local context "{index_name}", attempting lookup...')
        try:
            # Ensure we're using the index UUID, lookup by name is no longer
            # possible
            uuid.UUID(index_name)
            index_uuid = index_name

            click.secho('looking up {}...'.format(index_uuid))
            sc = pc.get_search_client()
            index_info = sc.get_index(index_uuid)
            display_name = index_info['display_name']
            pc.context.add_context(display_name, {
                'client_id': 'e4d82438-00df-4dbd-ab90-b6258933c335',
                'app_name': '{} app'.format(display_name),
                'manifest_index': index_uuid,
                'manifest_subject': 'globus://project-manifest.json',
                'scopes': pc.DEFAULT_SCOPES,
                'projects_cache_timeout': DEFAULT_PROJECTS_CACHE_TIMEOUT,
                'projects_endpoint': '',
                'projects_base_path': '',
                'projects_group': '',
                'projects_default_search_index': index_uuid,
                'projects_default_resource_server': 'petrel_https_server',
            })
            index_name = display_name
        except ValueError:
            click.secho(f'"{index_name}": must be a UUID when adding a new '
                        f'un-tracked search index.', fg='red')
            sys.exit(1)
        except Exception as e:
            log.exception(e)
            click.secho('Unable to find index {}'.format(index_name))
            sys.exit(2)
    pc.context.set_context(index_name)
    try:
        log.debug('Updating index...')
        pc.context.update()
    except exc.NoManifestException:
        click.secho('Pilot has not been setup for this index. Run '
                    '"pilot index setup" to do so.', fg='red')
        sys.exit(3)

    # If there is only one project, automatically set pilot to that one
    if len(pc.project.load_all()) == 1:
        projects = pc.project.load_all()
        pc.project.current = list(projects.keys())[0]
    log.debug('Context set! Fetching general info for user...')
    ctx.invoke(info, index=index_name)
    ctx.invoke(project_command)
    try:
        pc.load_tokens(requested_scopes=pc.context.get_value('scopes'))
    except ScopesMismatch:
        click.secho('Scopes do not match for this index, please login '
                    'again.', fg='red')
    log.debug(pc.context.get_value('projects_default_search_index'))
    if not pc.project.load_all():
        example = '''
        [[default-project]]
        base_path = /
        description =
        endpoint =
        group =
        resource_server = petrel_https_server
        search_index = {}
        title = default-project
        '''.format(pc.context.get_value('projects_default_search_index'))
        click.secho('No projects in this index, you can bootstrap one in the '
                    'config under [projects]\n{}'.format(example))


@index_command.command(help='Print Info for index')
@click.argument('index', required=False)
def info(index=None):
    pc = commands.get_pilot_client()
    index = index or pc.context.current
    if index is None:
        click.echo('Use "pilot index info <context>" to list info about a '
                   'project.')
        return
    try:
        info = pc.context.get_context(index)
    except exc.PilotInvalidProject as pip:
        click.secho(str(pip), fg='red')
        return

    fmt = '{:40.39}{}'
    log.debug(info)
    output = '\n'.join([fmt.format(*i) for i in info.items()])
    click.echo(output)
    click.echo()


@index_command.command(help='Setup pilot on a search index')
def setup():
    """Setup a brand new pilot config for a new index.

    NOTE: Paths with tilde '~' DO NOT WORK. These cause problems for resolving
    paths that look like /~/foo/bar, which sometimes translate as ~/foo/bar
    instead. These are disabled to prevent that from happening.
    """
    PROJECT_QUERIES = {
        'projects_endpoint': {
            'prompt': 'Set a Globus UUID where your data should be stored.',
            'default': '',
            'help': 'visit "https://app.globus.org/file-manager/collections" '
                    'to find Globus endpoint to store your data.',
            'validation': [input_validation.validate_is_uuid,
                           input_validation.validate_is_globus_endpoint],
        },
        'projects_base_path': {
            'prompt': 'Pick a base path.',
            'default': '/',
            'help': 'All data will be saved under this directory',
            'validation': [input_validation.validate_no_spaces,
                           input_validation.validate_absolute_path,
                           input_validation.validate_no_tilde],
        },
        'projects_group': {
            'prompt': 'Pick a Globus Group to secure Globus Search records',
            'default': 'public',
            'help': 'The group determines who can view records in search. '
                    'People not in this group will not see records in Globus '
                    'Search. "public" allows anyone to see these records.',
            'validation': [input_validation.validate_is_valid_globus_group],
        },
    }
    pc = commands.get_pilot_client()
    projects = pc.project.load_all().keys()
    if projects:
        click.secho(
            f'Index is already setup with the following projects: {projects}. '
            f'Please delete them before setting up your index.', fg='red')
        return
    order = ['projects_endpoint', 'projects_base_path', 'projects_group']

    iv = input_validation.InputValidator(queries=PROJECT_QUERIES, order=order)
    new_ctx = iv.ask_all()
    pc.context.update_context(new_ctx)
    pc.context.push()
    click.secho('Your index has been setup successfully. Now see '
                '`pilot project add`.', fg='green')
    return


@index_command.command(help='Update a context')
def push():
    pc = commands.get_pilot_client()
    pc.context.push()
    click.secho('Global projects have been updated. Users will be notified '
                'within 24 hours.', fg='green')



import logging
from pilot import config
from pilot.exc import PilotInvalidProject

log = logging.getLogger(__name__)


class Project(config.ConfigSection):

    SECTION = 'project'

    def load_all(self):
        return self.config.load().get('projects', {})

    def load_groups(self):
        return dict(self.config.load().get('groups', {}))

    def get_info(self, project=None):
        if project is None:
            project = self.current
        pinfo = self.load_all().get(project)
        if pinfo is None:
            raise PilotInvalidProject(f'No project exists {project}')
        return pinfo

    def is_set(self):
        """Returns true if a project has been set, false otherwise"""
        try:
            return self.current and self.current in self.load_all().keys()
        except PilotInvalidProject:
            return False

    def set_project(self, slug):
        self.current = slug

    def add_project(self, slug, project_data):
        cfg = self.config.load()
        cfg['projects'][slug] = project_data
        cfg.write()

    def delete_project(self, slug):
        if self.current == slug:
            self.current = None
        cfg = self.config.load()
        del cfg['projects'][slug]
        cfg.write()

    def lookup_group(self, group):
        reverse_lookup = {v: k for k, v in self.load_groups().items()}
        return reverse_lookup.get(group)

    def purge(self):
        cfg = self.config.load()
        cfg['projects'] = {}
        cfg['groups'] = {}
        cfg['project']['current'] = ''
        cfg.write()

    @property
    def current(self):
        curr = self.load_option('current')
        if curr is None:
            raise PilotInvalidProject('No current project configured')
        return curr

    @current.setter
    def current(self, value):
        projects = list(self.load_all().keys())
        if value not in projects and value is not None:
            raise ValueError(f'Project must be one of: {", ".join(projects)}')
        self.save_option('current', value)



from click.testing import CliRunner
from pilot.commands.search.delete import delete_command


def test_delete(mock_cli):
    runner = CliRunner()
    mock_cli.ls.return_value = {'bar': {'type': 'file'}}
    result = runner.invoke(delete_command, ['foo/bar'])
    assert result.exit_code == 0
    assert mock_cli.delete_entry.called



from PIL import Image


def analyze_image(filename, foreign_keys=None):
    img = Image.open(filename)
    image_metadata = {
        'name': 'Image Metadata',
        'width': img.width,
        'height': img.height,
        'colorspace': img.mode,
        'format': img.format,
        'format_description': img.format_description,
        }
    return image_metadata



import pytest
from urllib.parse import urlparse
from pilot.client import PilotClient
from pilot.exc import PilotInvalidProject

from tests.unit.mocks import MOCK_PROJECTS


def test_get_index(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    assert pc.get_index() == 'foo-search-index'
    assert pc.get_index(project='foo-project-test') == 'foo-test-search-index'


def test_invalid_project(mock_projects):
    pc = PilotClient()
    with pytest.raises(PilotInvalidProject):
        pc.get_index()


def test_invalid_project_with_explicit_name(mock_projects):
    pc = PilotClient()
    with pytest.raises(PilotInvalidProject):
        pc.get_index('does-not-exist')


def test_build_short_path(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    assert pc.build_short_path('foo', '/') == 'foo'
    assert pc.build_short_path('/foo/', '/') == 'foo'

    assert pc.build_short_path('foo', 'bar') == 'bar/foo'
    assert pc.build_short_path('/foo/', '/bar/') == 'bar/foo'


def test_get_path(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    assert pc.get_path('folder/file.txt') == '/foo_folder/folder/file.txt'
    path = pc.get_path('folder/file.txt', project='foo-project-test')
    assert path == '/foo_test_folder/folder/file.txt'


def test_special_paths(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    assert pc.get_path('///') == '/foo_folder'
    assert pc.get_path('.') == '/foo_folder'
    assert pc.get_path('..') == '/foo_folder'
    assert pc.get_path('/foo/bar/baz.txt') == '/foo_folder/foo/bar/baz.txt'


def test_get_globus_http_url(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    url = pc.get_globus_http_url('foo.txt')
    purl = urlparse(url)
    foo = MOCK_PROJECTS['foo-project']
    assert purl.netloc == foo['endpoint'] + '.e.globus.org'
    assert purl.scheme == 'https'
    assert purl.path == '/foo_folder/foo.txt'


def test_get_globus_url(mock_projects):
    foo = MOCK_PROJECTS['foo-project']
    pc = PilotClient()
    pc.project.current = 'foo-project'
    url = pc.get_globus_url('metadata/foo.txt')
    purl = urlparse(url)
    assert purl.netloc == foo['endpoint']
    assert purl.scheme == 'globus'
    assert purl.path == '/foo_folder/metadata/foo.txt'


def test_get_globus_app_url(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    url = pc.get_globus_app_url('metadata/foo.txt')
    purl = urlparse(url)
    assert purl.netloc == 'app.globus.org'
    assert purl.scheme == 'https'
    assert purl.path == '/file-manager'
    assert purl.query == 'origin_id=foo-project-endpoint&' \
                         'origin_path=%2Ffoo_folder%2Fmetadata%2Ffoo.txt'


def test_get_subject_url(mock_projects):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    args = ('myfolder/dataframe.dat',)
    assert pc.get_globus_url(*args) == pc.get_subject_url(*args)


def test_get_portal_url(mock_projects, mock_context):
    pc = PilotClient()
    pc.project.current = 'foo-project'
    assert pc.get_portal_url('') == 'https://myportal/foo-project/'
    assert pc.get_portal_url('foo') == (
        'https://myportal/foo-project/'
        'globus%253A%252F%252Ffoo-project-endpoint%252Ffoo_folder%252Ffoo/'
    )
    cfg = mock_context.load()
    del cfg['contexts']['test-context']['projects_portal_url']
    mock_context.save(cfg)
    assert pc.get_portal_url('foo') is None


def test_get_short_path_valid_urls(mock_cli, mock_paths):
    short_path = mock_paths['short_path']
    for path in mock_paths.values():
        assert mock_cli.get_short_path(path) == short_path

    # This WORKS, even though its the wrong project, since we can't
    # resolve the base path for the project
    bar_path = mock_cli.get_path(short_path, project='bar-project')
    assert bar_path.endswith(short_path)


def test_get_short_path_invalid_urls(mock_cli):
    short_path = 'test_path'
    invalid = [
        f'ftp://{mock_cli.get_endpoint(project="bar-project")}/{short_path}',
        mock_cli.get_subject_url(short_path, project='bar-project'),
        mock_cli.get_globus_http_url(short_path, project='bar-project'),
    ]
    for i in invalid:
        with pytest.raises(PilotInvalidProject):
            mock_cli.get_short_path(i)


def test_resolve_project(mock_cli, mock_paths):
    paths = mock_paths['short_path'], mock_paths['full_path']
    urls = mock_paths['subject'], mock_paths['http']
    for path in paths:
        assert mock_cli.resolve_project(path) is None
    for url in urls:
        assert mock_cli.resolve_project(url) == mock_cli.get_project()


def test_resolve_context(mock_cli, mock_paths):
    paths = mock_paths['short_path'], mock_paths['full_path']
    urls = mock_paths['subject'], mock_paths['http']
    for path in paths:
        assert mock_cli.resolve_context(path) is None
    for url in urls:
        assert mock_cli.resolve_context(url) == mock_cli.get_context()



import os
from unittest.mock import Mock
from pilot import search_parse
from tests.unit.mocks import ANALYSIS_FILE_BASE_DIR

test_search_record = os.path.join(ANALYSIS_FILE_BASE_DIR, 'dataset',
                                  'valid-typical.json')


def test_parse_record(monkeypatch, mock_search_data):
    log = Mock()
    monkeypatch.setattr(search_parse, 'log', log)
    data = search_parse.parse_result(mock_search_data)
    for info in data:
        assert all(info)
    print(f'Exceptions when parsing: {log.exception.call_args}')
    assert not log.exception.called


def test_parse_field_metadata(monkeypatch, mock_search_data):
    log = Mock()
    monkeypatch.setattr(search_parse, 'log', log)
    meta = mock_search_data['files'][0]['field_metadata']
    field_meta = search_parse.get_field_metadata(meta)
    both = ['name', 'type', 'count']
    number = ['min', 'max', 'mean', 'std', '25', '50', '75']
    string = ['count', 'frequency', 'top', 'unique']
    for entry in field_meta:
        fields = dict(entry)
        assert all([fields[i] for i in both])
        if fields['type'] == 'string':
            assert all(fields[i] for i in string)
        elif fields['type'] == 'float64':
            assert all(fields[i] for i in number)
    print(f'Exceptions when parsing: {log.exception.call_args}')
    assert not log.exception.called



import os
from pilot.search import (update_metadata, scrape_metadata,
                          get_files, prune_files, get_subdir_paths,
                          carryover_old_file_metadata)
from tests.unit.mocks import ANALYSIS_FILE_BASE_DIR, MULTI_FILE_DIR

MIXED_FILE = os.path.join(ANALYSIS_FILE_BASE_DIR, 'mixed.tsv')
NUMBERS_FILE = os.path.join(ANALYSIS_FILE_BASE_DIR, 'numbers.tsv')


def test_scrape_metadata(mock_cli):
    # dataframe, url, profile, project, skip_analysis = True
    pf = mock_cli.profile
    meta = scrape_metadata(MIXED_FILE, 'https://foo.com', pf, 'foo-project')
    dc_content = ['titles', 'creators', 'subjects', 'publicationYear',
                  'publisher', 'resourceType', 'dates', 'formats', 'version']

    assert all([c in meta['dc'].keys() for c in dc_content])
    assert meta['dc']['formats'] == ['text/tab-separated-values']
    assert meta['dc']['version'] == '1'
    assert meta['dc']['creators'][0]['creatorName'] == 'Franklin, Rosalind'
    assert set(meta.keys()) == {'dc', 'files', 'project_metadata'}
    assert 'field_metadata' in meta['files'][0].keys()
    assert 'foo-project' == meta['project_metadata']['project-slug']


def test_update_metadata_new_record_w_meta(mock_cli):
    pf = mock_cli.profile
    rec = scrape_metadata(MIXED_FILE, 'https://foo.com', pf, 'foo-project')
    meta = update_metadata(rec, None, {'mime_type': 'csv'})
    assert meta['dc']['formats'] == ['csv']


def test_update_metadata_new_file(mock_cli):
    pf = mock_cli.profile
    old = scrape_metadata(MIXED_FILE, 'globus://foo.com', pf, 'foo-project')
    new = scrape_metadata(NUMBERS_FILE, 'globus://bar.com', pf, 'foo-project')
    meta = update_metadata(new, old, {})
    assert old['files'][0] in meta['files']
    assert new['files'][0] in meta['files']


def test_prune_files(mock_multi_file_result):
    entry = mock_multi_file_result['gmeta'][0]['content'][0]
    assert len(entry['files']) == 4
    assert len(prune_files(entry,
                           '/foo_folder/multi_file/text_metadata.txt')) == 3
    assert len(prune_files(entry,
                           '/foo_folder/multi_file/')) == 0
    assert len(prune_files(entry,
                           '/foo_folder/multi_file/folder/')) == 1
    assert len(prune_files(entry,
                           '/foo_folder/multi_file/folder/folder2')) == 3

    assert len(prune_files(entry,
                           '/foo_folder/multi_file/folder')) == 1


def test_update_metadata_prev_record(mock_cli, mock_profile):
    pf = mock_cli.profile
    old = scrape_metadata(MIXED_FILE, 'globus://foo.com', pf, 'foo-project')
    mock_cli.profile.name = 'Marie Curie'
    new = scrape_metadata(MIXED_FILE, 'globus://foo.com', pf, 'foo-project')

    assert new != old
    assert old['dc']['creators'][0]['creatorName'] == 'Franklin, Rosalind'

    meta = update_metadata(new, old, {})
    assert meta['dc']['creators'][0]['creatorName'] == 'Curie, Marie'
    assert meta['files'] == new['files'] == old['files']


def test_carryover_old_file_metadata_same_file():
    file1 = [{'url': 'foo.txt'}]
    assert carryover_old_file_metadata(file1, file1) == file1


def test_carryover_old_file_metadata_diff_mimetype():
    old = [{'url': 'foo.txt', 'mime_type': 'text/plain'}]
    new = [{'url': 'foo.txt'}]
    assert carryover_old_file_metadata(new, old) == old


def test_carryover_old_file_metadata_skips_attrs():
    old = [{'url': 'foo.txt', 'length': 10, 'md5': 'abc', 'sha256': 'xyz'}]
    new = [{'url': 'foo.txt', 'length': 20, 'md5': 'def'}]
    assert carryover_old_file_metadata(new, old) == new


def test_carryover_old_file_metadata_does_not_overwrite():
    old = [{'url': 'foo.txt', 'mime_type': 'text/csv'}]
    new = [{'url': 'foo.txt', 'mime_type': 'text/plain'}]
    assert carryover_old_file_metadata(new, old) == new


def test_carryover_old_file_metadata_with_mix():
    old = [{'url': 'foo.txt', 'length': 10, 'md5': 'abc', 'sha256': 'xyz',
            'custom_attr': 'foo', 'extra_attr': 'bar'}]
    new = [{'url': 'foo.txt', 'length': 20, 'md5': 'def',
            'custom_attr': 'foo'}]
    expected = [{'url': 'foo.txt', 'length': 20, 'md5': 'def',
                 'custom_attr': 'foo', 'extra_attr': 'bar'}]
    assert carryover_old_file_metadata(new, old) == expected


def test_carryover_unrelated_files():
    old = [{'url': 'foo'}]
    new = [{'url': 'bar'}]
    assert carryover_old_file_metadata(new, old) == new + old


def test_update_file_version(mock_cli, mock_profile):
    pf = mock_cli.profile
    ver_one = scrape_metadata(MIXED_FILE, 'globus://foo.com', pf,
                              'foo-project')
    ver_two = scrape_metadata(NUMBERS_FILE, 'globus://foo.com', pf,
                              'foo-project')
    assert ver_one['dc']['version'] == '1'
    assert ver_two['dc']['version'] == '1'
    # Pretend 'NUMBERS_FILE' is an updated version of MIXED_FILE
    updated = update_metadata(ver_two, ver_one, {})
    assert updated['dc']['version'] == '2'
    # Pretend we switched back
    updated_3 = update_metadata(ver_one, updated, {})
    assert updated_3['dc']['version'] == '3'
    # Check re-updating does not bump version
    updated_3 = update_metadata(ver_one, updated, {})
    assert updated_3['dc']['version'] == '3'


def test_get_files():
    files = get_files(MULTI_FILE_DIR)
    assert len(files) == 4
    one_file = get_files(MIXED_FILE)
    assert len(one_file) == 1


def test_get_subdir_paths_on_dir():
    for local_path, remote_path in get_subdir_paths(MULTI_FILE_DIR):
        assert MULTI_FILE_DIR in local_path
        folder_name = os.path.basename(MULTI_FILE_DIR)
        assert remote_path.startswith(folder_name)


def test_get_subdir_paths_on_file():
    for local_path, remote_path in get_subdir_paths(MIXED_FILE):
        assert os.path.basename(local_path) == remote_path



from globus_sdk.exc import GlobusAPIError
from enum import IntEnum


class ExitCodes(IntEnum):
    SUCCESS = 0
    UNCAUGHT_EXCEPTION = 1
    NOT_LOGGED_IN = 2
    NO_DESTINATION_PROVIDED = 3
    DIRECTORY_DOES_NOT_EXIST = 4
    GLOBUS_TRANSFER_ERROR = 5
    INVALID_METADATA = 6
    RECORD_EXISTS = 7
    INVALID_CLIENT_CONFIGURATION = 8
    NO_LOCAL_ENDPOINT_SET = 9
    LOCAL_ENDPOINT_UNRESPONSIVE = 10
    DESTINATION_IS_RECORD = 11
    NO_RECORD_EXISTS = 12
    INVALID_DATAFRAME_NAME = 13


class PilotClientException(Exception):
    pass


class InvalidProject(PilotClientException):
    pass


class SubjectOutsideProject(PilotClientException):
    pass


class PilotContextException(Exception):
    pass


class NoManifestException(PilotContextException):
    pass


class PilotInvalidProject(PilotClientException):
    pass


class PilotValidator(PilotClientException):

    def __init__(self, message, *args, **kwargs):
        super().__init__()
        self.message = message or 'Error Validating Input'

    def __str__(self):
        return self.message


class RecordDoesNotExist(PilotClientException):
    pass


class PilotCodeException(PilotClientException):
    """Pilot Code Exceptions are a general class for any exception that might
    be thrown during the execution of a pilot command. The main difference from
    regular exceptions is a CODE, which must correspond to exc.ExitCodes, which
    is the integer which will be passed to sys.exit(). This is solely to
    facilitate bash scripting, so someone can check the exit code of a command
    and have enough context to make a decision with that information.

    Calling str(pce) on these exceptions must yield a user readable error,
    where repr(pce) might also yield the exit code.
    """
    MESSAGE = 'Unknown Error'
    CODE = ExitCodes.UNCAUGHT_EXCEPTION

    def __init__(self, message=None, fmt=None, verbose=False):
        super().__init__()
        self.message = message or self.MESSAGE
        self.fmt = fmt
        if fmt:
            self.message = self.message.format(*fmt)
        self.verbose = verbose
        self.verbose_output = ''

    def __str__(self):
        return '{}\n{}'.format(
            self.message, self.verbose_output if self.verbose else '')

    def __repr__(self):
        return '({}) {}\n{}'.format(
            self.CODE.name, self.message,
            self.verbose_output if self.verbose else '')


class NoDestinationProvided(PilotCodeException):
    MESSAGE = ('No Destination Provided. Please select one from the '
               'directory or "/" for root:\n{}')
    CODE = ExitCodes.NO_DESTINATION_PROVIDED


class InvalidDataframeName(PilotCodeException):
    MESSAGE = 'Dataframe may not be empty or start with symbols ("/" ok): {}'
    CODE = ExitCodes.INVALID_DATAFRAME_NAME


class DirectoryDoesNotExist(PilotCodeException):
    MESSAGE = 'Directory does not exist: "{}"'
    CODE = ExitCodes.DIRECTORY_DOES_NOT_EXIST


class GlobusTransferError(PilotCodeException):
    MESSAGE = '{}'
    CODE = ExitCodes.GLOBUS_TRANSFER_ERROR


class NoChangesNeeded(PilotCodeException):
    MESSAGE = '"{}": Files and search entry are an exact match.'
    CODE = ExitCodes.SUCCESS


class DestinationIsRecord(PilotCodeException):
    MESSAGE = ('The Destination "{}" is a record. Adding records to existing '
               'collections is not supported.')
    CODE = ExitCodes.DESTINATION_IS_RECORD


class RecordExists(PilotCodeException):
    MESSAGE = '"{}": Record Exists, extra confirmation needed to overwrite.'
    CODE = ExitCodes.RECORD_EXISTS

    def __init__(self, previous_metadata, fmt=None, verbose=False):
        super().__init__(verbose=verbose, fmt=fmt)
        self.previous_metadata = previous_metadata


class DryRun(PilotCodeException):
    MESSAGE = 'Success! (Dry Run -- No changes applied.)'
    CODE = ExitCodes.SUCCESS

    def __init__(self, stats=None, verbose=False):
        super().__init__(verbose=verbose)
        self.message = self.MESSAGE
        self.stats = stats
        self.previous_metadata = stats.get('previous_metadata')
        self.new_metadata = stats['new_metadata']
        self.verbose_output = self.stats, self.new_metadata


class NoLocalEndpointSet(PilotCodeException):
    MESSAGE = ('No Local endpoint set. If you are running locally and GCP is '
               'installed, start GCP then run `pilot login --force` for pilot '
               'to autodetect your endpoint. If you are running on GCS, you '
               'can use `pilot profile --local-endpoint UUID:PATH`')
    CODE = ExitCodes.NO_LOCAL_ENDPOINT_SET


class LocalEndpointUnresponsive(PilotCodeException):
    MESSAGE = ('Your local endpoint is not responding. Is GCP Running? '
               'Transfer Response: {}')
    CODE = ExitCodes.LOCAL_ENDPOINT_UNRESPONSIVE


class InvalidField(PilotClientException):
    pass


class AnalysisException(PilotClientException):
    def __init__(self, message, exc):
        super().__init__(message)
        self.message = message
        self.original_exc_info = exc


class HTTPSClientException(PilotClientException, GlobusAPIError):
    pass



import time
import datetime
import logging

from pilot import config

log = logging.getLogger(__name__)


class TransferLog(config.ConfigSection):

    SECTION = 'transfer_log'
    TRANSFER_LOG_FIELDS = ['dataframe', 'status', 'task_id', 'start_time']

    def _save_log(self, log_id, log_dict):
        if isinstance(log_dict['start_time'], datetime.datetime):
            timestamp = log_dict['start_time'].timestamp()
            log_dict['start_time'] = str(int(timestamp))
        field_list = [log_dict.get(f) for f in self.TRANSFER_LOG_FIELDS]
        log_data = ','.join(field_list)
        self.save_option(str(log_id), log_data)

    def add_log(self, transfer_result, datapath):
        cfg = self.config.load()
        last_id = max([int(i) for i in cfg['transfer_log'].keys()] or [-1])
        log_id = str(last_id + 1)
        log_data = [
            datapath,
            transfer_result.data['code'],
            transfer_result.data['task_id'],
            str(int(time.time()))
        ]
        self._save_log(log_id, dict(zip(self.TRANSFER_LOG_FIELDS, log_data)))
        log.debug('Log saved successfully.')

    def get_log(self):
        cfg = self.config.load()
        if 'transfer_log' not in cfg:
            return []

        logs = []
        for log_id, data in dict(cfg['transfer_log']).items():
            tlog = dict(zip(self.TRANSFER_LOG_FIELDS, data.split(',')))
            tlog['id'] = int(log_id)
            timestamp = int(tlog['start_time'])
            tlog['start_time'] = datetime.datetime.fromtimestamp(timestamp)
            logs.append(tlog)
        logs.sort(key=lambda l: l['id'], reverse=True)
        return logs

    def get_log_by_task(self, task_id):
        for tlog in self.get_log():
            if tlog['task_id'] == task_id:
                return tlog

    def update_log(self, task_id, new_status):
        tlog = self.get_log_by_task(task_id)
        tlog['status'] = new_status
        self._save_log(tlog['id'], tlog)



import os
from unittest.mock import Mock
from click.testing import CliRunner
from pilot.commands.auth.auth_commands import (
    login, logout, profile_command)


def test_auth_login(monkeypatch, mock_cli, mock_config, mock_transfer_client,
                    mock_sdk_response):
    monkeypatch.setattr(mock_cli, 'is_logged_in', Mock(return_value=False))
    mock_sdk_response.data = {'display_name': 'my_computer'}
    mock_transfer_client.get_endpoint.return_value = mock_sdk_response
    runner = CliRunner()
    result = runner.invoke(login, [])
    assert result.exit_code == 0
    assert mock_cli.login.called


def test_auth_logout(mock_cli):
    runner = CliRunner()
    result = runner.invoke(logout, [])
    assert result.exit_code == 0
    assert mock_cli.logout.called


def test_auth_logout_purge(monkeypatch, mock_cli):
    monkeypatch.setattr(os, 'unlink', Mock())
    monkeypatch.setattr(os.path, 'exists', Mock(return_value=True))
    runner = CliRunner()
    result = runner.invoke(logout, ['--purge'])
    assert result.exit_code == 0
    assert mock_cli.logout.called
    assert os.unlink.called


def test_auth_profile(mock_cli, mock_config):
    runner = CliRunner()
    result = runner.invoke(profile_command, [])
    assert result.exit_code == 0



import click

from pilot.commands import get_pilot_client
import globus_sdk

INACTIVE_STATES = ['SUCCEEDED', 'FAILED', 'CANCELED']


def update_tasks(transfer_tasks):
    """
    Update pending Globus Transfer tasks, and save the resulting status to
    the config. transfer tasks is a list of dicts as returned by
    config.get_pc.transfer_log()

    User must be logged in!
    """
    pc = get_pilot_client()
    auth = pc.get_authorizers()['transfer.api.globus.org']
    tc = globus_sdk.TransferClient(authorizer=auth)
    user_tasks = {r.data['task_id']: r.data for r in
                  tc.task_list(num_results=100).data}
    for task in transfer_tasks:
        task_data = user_tasks.get(task['task_id'])
        if not task_data:
            click.secho('Unable to update status for {}'.format(task['id']),
                        fg='yellow')
        else:
            status = task_data.get('nice_status') or task_data.get('status')
            pc.transfer_log.update_log(task['task_id'], status)


@click.command(help='Check status of transfers', name='status')
# @click.argument('task', required=False)
@click.option('-n', 'number', type=int, default=10,
              help='Number of tasks to list')
def status_command(number):
    pc = get_pilot_client()

    ordered_tlogs = []
    tlog_order = ['id', 'dataframe', 'status', 'start_time', 'task_id']
    # Fetch a limmited set of logs by the most recent entries
    tlogs = pc.transfer_log.get_log()[:number]

    pending_tasks = [t for t in tlogs if t['status'] not in INACTIVE_STATES]
    if pending_tasks:
        click.secho('Updating tasks...', fg='green')
        update_tasks(pending_tasks)
        tlogs = pc.transfer_log.get_log()[:number]

    for tlog in tlogs:
        tlog['id'] = str(tlog['id'])
        tlog['start_time'] = tlog['start_time'].strftime('%Y-%m-%d %H:%M')
        ordered_tlogs.append([tlog[item] for item in tlog_order])

    fmt = '{:4.3}{:30.29}{:10.11}{:18.17}{:37.36}'
    header_names = ['ID', 'Dataframe', 'Status', 'Start Time', 'Task ID']
    headers = fmt.format(*header_names)
    output = '\n'.join([fmt.format(*items) for items in ordered_tlogs])

    click.echo(headers)
    click.echo(output)



import logging
import sys
from pilot import exc
from pilot.analysis import mimetypes

log = logging.getLogger(__name__)


try:
    from pilot.analysis import pandas as pandalyze
    pandas_map = {
        'text/tab-separated-values': pandalyze.analyze_tsv,
        'text/csv': pandalyze.analyze_csv,
        'application/x-hdf': pandalyze.analyze_hdf,
        'application/x-parquet': pandalyze.analyze_parquet,
        'application/x-feather': pandalyze.analyze_feather,
    }
except ImportError:
    log.debug('Dependency not found, pandas analysis disabled',
              exc_info=True)
    pandas_map = {}

try:
    from pilot.analysis import image as imaginalyze
    image_map = {
        'image/jpeg': imaginalyze.analyze_image,
        'image/png': imaginalyze.analyze_image,
    }
except ImportError:
    log.debug('Dependency not found, imaging analysis disabled',
              exec_info=True)
    image_map = {}


ANALYZE_MAP = {}
ANALYZE_MAP.update(pandas_map)
ANALYZE_MAP.update(image_map)


def analyze_dataframe(filename, mimetype=None):
    mimetype = mimetype or mimetypes.detect_type(filename)
    analyze_function = ANALYZE_MAP.get(mimetype)
    if analyze_function is None:
        log.debug('No analyzer for mimetype {}'.format(mimetype))
        return {}
    try:
        return analyze_function(filename)
    except Exception as e:
        log.exception(e)
        log.error('Failed to parse metadata.')
        msg = 'Failed to analyze {}'.format(filename)
        raise exc.AnalysisException(msg, sys.exc_info()) from None



import pytest
from configobj import ConfigObj
import os
import json
import copy
import time
import globus_sdk
from unittest.mock import Mock
from .mocks import (MemoryStorage, MOCK_TOKEN_SET, GlobusTransferTaskResponse,
                    GlobusResponse,
                    ANALYSIS_FILE_BASE_DIR, SCHEMA_FILE_BASE_DIR,
                    CLIENT_FILE_BASE_DIR,
                    MOCK_PROFILE, MOCK_PROJECTS, MOCK_CONTEXT)

from pilot import client, config, commands, transfer_log


@pytest.fixture
def mem_storage():
    return MemoryStorage()


@pytest.fixture
def mock_tokens():
    return copy.deepcopy(MOCK_TOKEN_SET)


@pytest.fixture
def mock_config(monkeypatch):
    class MockConfig:
        cfg = ConfigObj()

        def save(self, cfg):
            self.cfg = cfg

        def load(self):
            return self.cfg

    mock_cfg = MockConfig()
    monkeypatch.setattr(config.Config, 'load', mock_cfg.load)
    monkeypatch.setattr(config.Config, 'save', mock_cfg.save)
    return config.Config('/tmp/test_pilot.cfg')


@pytest.fixture
def mock_profile(mock_config):
    cfg = mock_config.load()
    cfg['profile'] = MOCK_PROFILE
    mock_config.save(cfg)
    return mock_config


@pytest.fixture
def mock_projects(mock_config, mock_contexts):
    cfg = mock_config.load()
    cfg['projects'] = MOCK_PROJECTS
    mock_config.save(cfg)
    return mock_config


@pytest.fixture
def mock_contexts(mock_config):
    cfg = mock_config.load()
    cfg['contexts'] = MOCK_CONTEXT
    cfg['context'] = {'current': 'test-context', 'last_updated': time.time()}
    mock_config.save(cfg)
    return mock_config


@pytest.fixture
def mock_context(mock_config):
    cfg = mock_config.load()
    cfg['contexts'] = MOCK_CONTEXT
    cfg['context'] = {'current': list(MOCK_CONTEXT.keys())[0]}
    mock_config.save(cfg)
    return mock_config


@pytest.fixture
def mixed_tsv():
    return os.path.join(ANALYSIS_FILE_BASE_DIR, 'mixed.tsv')


@pytest.fixture
def numbers_tsv():
    return os.path.join(ANALYSIS_FILE_BASE_DIR, 'numbers.tsv')


@pytest.fixture
def mock_search_data():
    fname = os.path.join(SCHEMA_FILE_BASE_DIR, 'dataset', 'valid-typical.json')
    with open(fname) as fh:
        return json.load(fh)


@pytest.fixture
def mock_search_results(mock_search_data):
    return {'@datatype': 'GSearchResult', '@version': '2017-09-01', 'count': 1,
            'gmeta': [{'@datatype': 'GMetaResult', '@version': '2017-09-01',
                       'content': [mock_search_data],
                       'subject': 'foo_folder'}],
            'offset': 0, 'total': 1}


@pytest.fixture
def mock_multi_file_result(mock_search_results):
    data = mock_search_results.copy()
    mf_filename = os.path.join(CLIENT_FILE_BASE_DIR, 'multi_file_entry.json')
    with open(mf_filename) as f:
        multi_file = json.load(f)
    sub = 'globus://foo-project-endpoint/foo_folder/multi_file'
    data['gmeta'] = [{'content': [multi_file], 'subject': sub}]
    return data


@pytest.fixture
def mock_search_result(mock_search_results):
    return mock_search_results['gmeta'][0]


@pytest.fixture
def mock_transfer_data(monkeypatch):
    td = Mock()
    monkeypatch.setattr(globus_sdk, 'TransferData', Mock(return_value=td))
    return td


@pytest.fixture
def mock_transfer_client(monkeypatch, mock_transfer_data):
    tc = Mock()
    monkeypatch.setattr(globus_sdk, 'TransferClient', Mock(return_value=tc))
    tc.submit_transfer.return_value = GlobusTransferTaskResponse()
    return tc


@pytest.fixture
def mock_search_client(monkeypatch):
    sc = Mock()
    gr = GlobusResponse()
    gr.data = {'state': 'SUCCESS', 'message': 'a thing has been ingested'}
    sc.get_task.return_value = gr
    sc.ingest.return_value = {'task_id': 'mock_task_id'}
    monkeypatch.setattr(globus_sdk, 'SearchClient', Mock(return_value=sc))
    return sc


@pytest.fixture
def mock_globus_exception():
    class MockExc(Exception):
        code = 'Error'
        message = 'A Globus SDK Transfer Error occurred! (Mock)'
    return MockExc


@pytest.fixture
def mock_transfer_error(monkeypatch, mock_globus_exception):
    monkeypatch.setattr(globus_sdk.exc, 'TransferAPIError',
                        mock_globus_exception)
    return mock_globus_exception


@pytest.fixture
def mock_sdk_response():
    return GlobusResponse()


@pytest.fixture
def mock_cli_basic(monkeypatch, mock_config, mock_projects, mock_contexts):
    pc = client.PilotClient()

    def load_tokens(*args, **kwargs):
        return MOCK_TOKEN_SET

    monkeypatch.setattr(pc, 'load_tokens', load_tokens)
    monkeypatch.setattr(commands, 'get_pilot_client', Mock(return_value=pc))
    monkeypatch.setattr(pc, 'config', mock_config)
    pc.config = mock_config
    pc.profile.config = mock_config
    pc.project.config = mock_config
    pc.project.current = 'foo-project'
    # Sanity. This *should* always return True, but will fail if we update
    # tokens at a later time.
    assert pc.is_logged_in()
    return pc


@pytest.fixture
def mock_transfer_log(monkeypatch):
    add_log = Mock()
    monkeypatch.setattr(transfer_log.TransferLog, 'add_log', add_log)
    return add_log


@pytest.fixture
def mock_cli(mock_cli_basic, mock_transfer_client, mock_search_client,
             mock_profile, mock_transfer_log):
    """
    Returns a mock logged in pilot client. Storage is mocked with a custom
    object, so this does behave slightly differently than the real client.
    All methods that reach out to remote resources are mocked, you need to
    re-mock them to return the test data you want.
    """
    mock_cli_basic.transfer_file = Mock()
    mock_cli_basic.login = Mock()
    mock_cli_basic.logout = Mock()
    mock_cli_basic.get_full_search_entry = Mock(return_value=None)
    mock_cli_basic.search = Mock(return_value={'gmeta': []})
    mock_cli_basic.ls = Mock()
    mock_cli_basic.mkdir = Mock()
    mock_cli_basic.delete = Mock()
    mock_cli_basic.delete_entry = Mock()
    mock_cli_basic.get_auth_client = Mock()
    mock_cli_basic.get_http_client = Mock()
    return mock_cli_basic


@pytest.fixture
def mock_paths(mock_cli):
    short_path = 'test_path'
    return {
        'short_path': short_path,
        'full_path': mock_cli.get_path(short_path),
        'subject': mock_cli.get_subject_url(short_path),
        'http': mock_cli.get_globus_http_url(short_path),
    }



import os
import json
import jsonschema

BASE_DIR = os.path.dirname(__file__)
BASE_SCHEMA_DIR = os.path.join(BASE_DIR, 'schemas')


def get_schemas():
    schemas = {}
    files = [f for f in os.listdir(BASE_SCHEMA_DIR)
             if os.path.splitext(f)[1] == '.json']

    for f in files:
        fname = os.path.join(BASE_SCHEMA_DIR, f)
        with open(fname) as fh:
            sname, _ = os.path.splitext(f)
            try:
                schemas[sname] = json.load(fh)
            except json.decoder.JSONDecodeError:
                raise Exception('Error loading {}'.format(fname))
    return schemas


def validate_dataset(dataset):
    validate_json('dataset', dataset)


def validate_user_provided_metadata(metadata):
    validate_json('user_provided_metadata', metadata)


def validate_json(name, json):
    schema = get_schemas()[name]
    resolver = jsonschema.RefResolver(
        base_uri="file://{}/{}".format(BASE_SCHEMA_DIR, name),
        referrer=name
    )
    jsonschema.validate(schema=schema, resolver=resolver, instance=json)



import os
import sys
import logging
import click
from slugify import slugify
import globus_sdk

from pilot import commands, exc
from pilot.commands import input_validation

log = logging.getLogger(__name__)

PROJECT_QUERIES = {
    'title': {
        'prompt': 'Pick a title for your new project',
        'default': 'My New Project',
        'help': 'Pick something short and easy to remember',
        'validation': [input_validation.validate_project_title_unique],
    },
    'short_name': {
        'prompt': 'Pick a short name, this will create a directory on "{}" to '
                  'store your files.',
        'default': lambda v, q: slugify(v.answers['title']),
        'help': 'The short name will be used in URLs and will be the name '
                'users select this new project',
        'validation': [input_validation.validate_no_spaces,
                       input_validation.validate_project_slug_unique,
                       input_validation.validate_slug_to_path_unique],
    },
    'description': {
        'prompt': 'Describe your new project',
        'default': 'This project is intended to do X for scientists',
        'help': 'A nice description can help people understand what your '
                'project does, at a glance',
        'validation': [],
    },
}


@click.group(name='project', help='Set or display project information',
             invoke_without_command=True)
@click.pass_context
def project_command(ctx):
    pc = commands.get_pilot_client()
    if not pc.is_logged_in():
        click.echo('You are not logged in.')
        return
    invalid_with_pending_update = ['delete', 'add']
    if ctx.invoked_subcommand in invalid_with_pending_update:
        if any(pc.context.update_with_diff(dry_run=True).values()):
            click.secho('There is an update for projects, please update '
                        '("pilot project update") before adding a new project',
                        fg='red')
            sys.exit()
    if ctx.invoked_subcommand is None:
        click.echo('Set project with "pilot project set <myproject>"')
        projects = pc.project.load_all()
        current = pc.project.current if pc.project.is_set() else None
        fmt = '{} {}'
        for project in projects:
            if project == current:
                click.secho(fmt.format('*', project), fg='green')
            else:
                click.echo(fmt.format(' ', project))


@project_command.command(help='Update stored list of projects.')
@click.option('--dry-run', is_flag=True, default=False)
@click.option('--update-groups-cache/--no-update-groups-cache', default=True,
              help='Fetch the latest subgroups from Globus')
def update(dry_run, update_groups_cache):
    pc = commands.get_pilot_client()
    try:
        output = pc.context.update_with_diff(
            dry_run=dry_run, update_groups_cache=update_groups_cache)
        if not any(output.values()):
            click.secho('Project is up to date', fg='green')
            return
        for group, changes in output.items():
            if not changes:
                continue
            click.echo('{}:'.format(group))
            for change_type, items in changes.items():
                click.echo('\t{}:'.format(change_type))
                for name, value in items.items():
                    click.echo('\t\t{} -> {}'.format(name, value))
    except globus_sdk.exc.SearchAPIError as sapie:
        click.secho(str(sapie), fg='red')
        click.secho('You can create the manifest for this index with `pilot '
                    'context push`', fg='blue')


@project_command.command(name='set', help='Set your project')
@click.argument('project', required=True)
def set_command(project):
    pc = commands.get_pilot_client()
    try:
        pc.project.current = project
        click.echo(f'Current project set to {project}')
    except ValueError as ve:
        click.secho(str(ve), fg='red')


@project_command.command(help='Add a new project')
def add():
    pc = commands.get_pilot_client()
    order = ['title', 'short_name', 'description']
    base_path = pc.context.get_value('projects_base_path')
    PROJECT_QUERIES['short_name']['prompt'] = \
        PROJECT_QUERIES['short_name']['prompt'].format(base_path)
    titles = [p['title'] for p in pc.project.load_all().values()]
    PROJECT_QUERIES['title']['current'] = titles

    iv = input_validation.InputValidator(queries=PROJECT_QUERIES, order=order)
    project = iv.ask_all()
    project.update(
        {'search_index': pc.context.get_value('projects_default_search_index'),
         'resource_server':
            pc.context.get_value('projects_default_resource_server')}
    )
    project['endpoint'] = pc.context.get_value('projects_endpoint')
    project['group'] = pc.context.get_value('projects_group')
    short_name = project.pop('short_name')
    project['base_path'] = os.path.join(base_path, short_name)

    click.secho('Updating global project list... ', nl=False)
    pc.project.add_project(short_name, project)
    tc = pc.get_transfer_client()
    tc.operation_mkdir(project['endpoint'], project['base_path'])
    pc.context.push()
    click.secho('Success', fg='green')
    pc.project.set_project(short_name)
    click.secho('Switched to project {}'.format(short_name))
    click.secho('Your new project "{}" has been added! '
                'Users will be notified within 24 hours next time they use '
                'this tool.'.format(project['title']), fg='green')


@project_command.command(help='Print project details')
@click.argument('project', required=False)
def info(project=None):
    pc = commands.get_pilot_client()
    current = pc.project.current if pc.project.is_set() else None
    project = project or current
    if project is None:
        click.echo('Use "pilot project info <project>" to list info about a '
                   'project.')
        return
    try:
        info = pc.project.get_info(project)
    except exc.PilotInvalidProject as pip:
        click.secho(str(pip), fg='red')
        return

    ep_name = info['endpoint']
    try:
        tc = pc.get_transfer_client()
        ep = tc.get_endpoint(info['endpoint']).data
        ep_name = ep['display_name'] or ep['canonical_name'] or ep_name
    except globus_sdk.exc.TransferAPIError:
        click.echo('Failed to lookup endpoint {}, please ensure it is active.'
                   .format(info['endpoint']))

    dinfo = [
        (info['title'], ''),
        ('Endpoint', ep_name),
        ('Group', pc.project.lookup_group(info['group'])),
        ('Base Path', info['base_path']),
    ]

    fmt = '{:25.24}{}'
    log.debug(info)
    output = '\n'.join([fmt.format(*i) for i in dinfo])
    click.echo(output)
    click.echo()
    click.echo(info['description'])


@project_command.command(help='Delete a project')
@click.option('--keep-context', is_flag=True, default=False,
              help='Continue to track the empty project in pilot after delete')
@click.argument('project', required=True)
def delete(project, keep_context):
    pc = commands.get_pilot_client()
    if project not in pc.project.load_all():
        click.secho('{} is not a valid project'.format(project), fg='red')
        return 1
    pc.project.current = project
    results = pc.search(project=project)
    pinfo = pc.project.get_info(project)
    search_query = {
        'q': '*',
        'filters': {
            'field_name': 'project_metadata.project-slug',
            'type': 'match_all',
            'values': [project or pc.project.current]
        }
    }
    project_base_path = pc.get_path('', project=project)
    search_client = pc.get_search_client()
    transfer_client = pc.get_transfer_client()
    log.debug('Base path for delete is: {}'.format(project_base_path))
    dz = '\n{}\nDANGER ZONE\n{}'.format('/' * 80, '/' * 80)
    click.secho('{dz}\n'
                'This will delete all data and search results in your '
                'project.\n{tot} datasets will be deleted for {project}\n\n'
                'Base Directory "{project_base_path}" will be deleted.'
                '{dz}'
                ''.format(dz=dz, tot=results['total'], project=project,
                          project_base_path=project_base_path),
                bg='red')
    click.echo('Please type the name ({}) of your project to delete it> '
               .format(project), nl=False)
    if input() != project:
        click.secho('Names do not match, aborting...')
        return 1
    click.echo(f'Deleting Data: {project_base_path}')
    try:
        ddata = globus_sdk.DeleteData(transfer_client, pinfo['endpoint'],
                                      recursive=True,
                                      notify_on_succeeded=False)
        ddata.add_item(project_base_path)
        transfer_client.submit_delete(ddata)
    except globus_sdk.exc.TransferAPIError as tapie:
        log.debug('Error deleting base folder', exc_info=tapie)
        click.secho(f'Error deleting project base folder {project_base_path}: '
                    f'{str(tapie)}', fg='red')

    try:
        click.echo(f'Deleting Search Records: {pinfo["search_index"]}')
        log.debug(f'Search Query: {search_query}')
        search_client.delete_by_query(pinfo['search_index'], search_query)
    except globus_sdk.exc.SearchAPIError as sapie:
        log.debug('Error deleting test data', exc_info=sapie)
        click.secho(f'Error deleting search data {pinfo["search_index"]}: '
                    f'{str(sapie)}', fg='red')
    if keep_context:
        click.secho('Keeping now empty pilot project {}'.format(project),
                    fg='blue')
        pc.mkdir('/')
    else:
        click.echo('Removing project...')
        pc.project.delete_project(project)
        pc.context.push()
    click.secho('Project {} has been deleted successfully.'.format(project),
                fg='green')


@project_command.command(help='Edit items about your project')
@click.argument('project', required=False)
def edit(project=None):
    pc = commands.get_pilot_client()
    current = pc.project.current if pc.project.is_set() else None
    project = project or current
    if project is None:
        click.echo('Use "pilot project info <project>" to list info about a '
                   'project.')
        return
    info = pc.project.get_info(project)
    queries = {'title': PROJECT_QUERIES['title'].copy(),
               'description': PROJECT_QUERIES['description'].copy()}
    titles = [p['title'] for p in pc.project.load_all().values()
              if p['title'] != info['title']]
    queries['title']['current'] = titles

    for key in queries:
        queries[key]['default'] = info[key]
    iv = input_validation.InputValidator(queries=queries,
                                         order=['title', 'description'])
    new_info = iv.ask_all()
    info.update(new_info)
    pc.project.set_project(project, info)
    pc.context.push()
    click.secho('Project "{}" has been updated'.format(project), fg='green')


@project_command.command(help='Update the global list of projects',
                         hidden=True)
def push():
    pc = commands.get_pilot_client()
    pc.context.push()
    click.secho('Global projects have been updated. Users will be notified '
                'within 24 hours.', fg='green')



from click.testing import CliRunner
from pilot.commands.search.search_commands import list_command, describe


def test_list_command(monkeypatch, mock_cli, mock_search_results):
    mock_cli.search.return_value = mock_search_results
    mock_cli.ls.return_value = {}
    runner = CliRunner()
    result = runner.invoke(list_command, [])
    assert result.exit_code == 0


def test_upload_gcp_log(mock_cli, mock_search_result):
    mock_cli.get_full_search_entry.return_value = mock_search_result
    runner = CliRunner()
    result = runner.invoke(describe, ['foo/bar'])
    assert result.exit_code == 0



import re
import logging
from globus_sdk import AccessTokenAuthorizer, RefreshTokenAuthorizer
from globus_sdk.base import BaseClient, slash_join
from globus_sdk.response import GlobusResponse
from pilot.exc import HTTPSClientException
from globus_sdk import exc
import requests
from requests_toolbelt import multipart

log = logging.getLogger(__name__)

# Pattern of how ranges must be given on the command line
RANGE = re.compile(r'^(\d+-\d+)(,\d+-\d+)*$')


class FileContentResponse(GlobusResponse):

    def __init__(self, response, client):
        super().__init__(response, client)

    @property
    def text(self):
        return str(self.data)

    @property
    def data(self):
        byte_content = b''
        for content in self.iter_content:
            byte_content += content
        return byte_content.decode('utf-8')

    @property
    def iter_content(self):
        content_type = self._data.headers.get('Content-Type', '')
        if 'multipart/byteranges' in content_type:
            md = multipart.decoder.MultipartDecoder.from_response(self._data)
            return [part.content for part in md.parts]
        return self._data.iter_content(chunk_size=(2**20 * 2))

    @property
    def raw_response(self):
        return self._data


class HTTPFileClient(BaseClient):
    allowed_authorizer_types = (AccessTokenAuthorizer, RefreshTokenAuthorizer)

    error_class = HTTPSClientException
    default_response_class = GlobusResponse
    file_content_response_class = FileContentResponse

    def __init__(self, authorizer=None, base_url='', **kwargs):
        super().__init__(
            self, "http_client", base_url=base_url,
            authorizer=authorizer, **kwargs
        )

    def get(self, path, params=None, headers=None, allow_redirects=False,
            filename=None, response_class=None, retry_401=True, range=None):
        headers = headers or {}
        if range:
            if not isinstance(range, str) or not re.match(RANGE, range):
                raise ValueError('Ranges must be a string of the pattern: '
                                 '"1-2,3-4" where each number is a byte range')
            headers['Range'] = 'bytes={}'.format(range)
        response_class = response_class or self.file_content_response_class
        return self.send_custom_request(
            'GET', path, params=params,
            headers=headers, allow_redirects=allow_redirects,
            response_class=response_class, retry_401=retry_401, stream=True
        )

    def put(self, path, params=None, headers=None, allow_redirects=False,
            filename=None, response_class=None, retry_401=True, data=None):
        if data:
            return self.send_custom_request(
                "PUT", path, params=params,
                headers=headers, allow_redirects=allow_redirects,
                response_class=response_class, retry_401=retry_401, data=data
            )
        if not filename:
            raise ValueError('No filename provided')
        with open(filename, 'rb') as data:
            self.logger.debug('PUT to {} with params {}'.format(path, params))
            return self.send_custom_request(
                "PUT", path, params=params,
                headers=headers, allow_redirects=allow_redirects, data=data,
                response_class=response_class, retry_401=retry_401,
            )

    def delete(self, path, **kwargs):
        return self.send_custom_request('DELETE', path, **kwargs)

    def send_custom_request(self, method, path, headers=None,
                            allow_redirects=False, response_class=None,
                            retry_401=True, **kwargs):
        rheaders = dict(self._headers)
        # expand
        if headers is not None:
            rheaders.update(headers)

        # add Authorization header, or (if it's a NullAuthorizer) possibly
        # explicitly remove the Authorization header
        if self.authorizer is not None:
            self.logger.debug(
                "request will have authorization of type {}".format(
                    type(self.authorizer)
                )
            )
            self.authorizer.set_authorization_header(rheaders)

        if path.startswith('https://') or path.startswith('http://'):
            url = path
        else:
            url = slash_join(self.base_url, path)
        self.logger.debug("request will hit URL:{}".format(url))

        # because a 401 can trigger retry, we need to wrap the retry-able thing
        # in a method
        def send_request():
            try:
                return self._session.request(
                    method=method,
                    url=url,
                    headers=rheaders,
                    allow_redirects=allow_redirects,
                    verify=self._verify,
                    timeout=self._http_timeout,
                    **kwargs
                )
            except requests.RequestException as e:
                self.logger.error("NetworkError on request")
                raise exc.convert_request_exception(e)

        # initial request
        r = send_request()

        self.logger.debug("Request made to URL: {}".format(r.url))

        # potential 401 retry handling
        if r.status_code == 401 and retry_401 and self.authorizer is not None:
            self.logger.debug("request got 401, checking retry-capability")
            # note that although handle_missing_authorization returns a T/F
            # value, it may actually mutate the state of the authorizer and
            # therefore change the value set by the `set_authorization_header`
            # method
            if self.authorizer.handle_missing_authorization():
                self.logger.debug("request can be retried")
                self.authorizer.set_authorization_header(rheaders)
                r = send_request()

        return self.handle_response(r, response_class)

    def handle_response(self, response, response_class=None):
        if 200 <= response.status_code < 400:
            self.logger.debug(
                "request completed with response code: {}".format(
                    response.status_code)
            )
            if response_class is None:
                return self.default_response_class(response, client=self)
            else:
                return response_class(response, client=self)

        self.logger.debug(
            "request completed with (error) response code: {}".format(
                response.status_code)
        )
        raise self.error_class(response)


class NexusClient(BaseClient):
    """This is a custom nexus client for fetching groups. It is not available
    outside of this client."""
    allowed_authorizer_types = (AccessTokenAuthorizer, RefreshTokenAuthorizer)
    error_class = HTTPSClientException
    default_response_class = GlobusResponse
    file_content_response_class = FileContentResponse

    def __init__(self, authorizer=None, base_url='', **kwargs):
        base_url = base_url or 'https://nexus.api.globusonline.org/groups/'
        super().__init__(
            self, "custom_pilot_nexus_client", base_url=base_url,
            authorizer=authorizer, **kwargs
        )

    def get_subgroups(self, group, **params):
        """
        Get a list of Globus Sub-Groups for a given group. Max depth is by
        default 100.
        """
        log.debug('Nexus GET for subgroups in "{}"'.format(group))
        request_params = {'root': group, 'depth': 100}
        request_params.update(params)
        return self.get('list', params=request_params)



import os
import urllib
import logging

log = logging.getLogger(__name__)


def get_sub_in_collection(subject, entries, precise=True):
    """
    Look for a subject in a bunch of search entries.
    If 'subject' refers to a file, this looks for an exact subject match in
    entries and returns the content if it exists. If 'subject' is a multi-file
    collection subject must either match the top level folder or a file within
    the collection (or a non-existent file within the collection if
    precise=False).
    **Parameters**
    ``subject`` (*string*)
      A Globus Search URL
    ``entries`` (*string*)
      A list of Globus Search GMeta entries.
    ``precise`` (*bool*)
      If the path given points to a location inside a multi-file directory
      only return the record if the location matches a file.
      For example, given an entry containing the files:
        my_dir/foo1.txt, my_dir/foo2.txt, my_dir/foo3.txt
      If precise=True and the path is my_dir/foo4.txt, None will be
      returned. If precise=False and the path is my_dir/foo4.txt, the
      "my_dir" record will still be returned.
    """
    entry = get_entry_with_matching_subject(entries, subject)
    if entry is None or precise is False:
        return entry
    urls = [m.get('url') for m in entry['content'][0].get('files', [])]
    sub_path = urllib.parse.urlparse(subject).path
    for url in urls:
        log.debug('Checking {}'.format(url))
        if sub_path in url:
            log.debug('Found specific file in entry: {}'.format(url))
            return entry


def get_entry_with_matching_subject(entries, subject):
    """
    Get an entry with a matching subject. A matching subject is both an exact
    match and a path to a file within the subject. For example, given the
    subject:
      globus://foo-project-endpoint/foo_subject
    These would all be valid matches:
      globus://foo-project-endpoint/foo_subject
      globus://foo-project-endpoint/foo_subject/foo.txt
      globus://foo-project-endpoint/foo_subject/bar.txt
      globus://foo-project-endpoint/foo_subject/sub_folder/foo.txt
    NOTE! The actual file inside a subject does not need to exist for a match
    to happen. globus://foo-project-endpoint/foo_subject/bar.txt will return
    a match whether or not 'bar.txt' exists in the 'files' manifest.
    """
    sub_map, directory = {ent['subject']: ent for ent in entries}, subject
    # If the given subject is a path within a search entry, this will derive
    # which subject its part of.
    while directory and directory not in sub_map.keys():
        directory = os.path.dirname(directory)

    matches = list(filter(lambda sub: sub == directory, sub_map.keys()))
    if len(matches) == 1:
        return sub_map[matches[0]]
    log.debug('Match Fail: Sub {} matched {}/{} subs, not 1: {}'
              ''.format(subject, len(matches), len(entries), matches))


def get_matching_file(url, entry):
    files = get_matching_files(url, entry)
    if len(files) == 1:
        return files[0]


def get_matching_files(url, entry):
    """Given a search entry returned from PilotClient.get_search_entry, find
    all of the partial matches in the entry. Partial matches can happen if
    a base folder matches, but not the files below it. For example:
      * A url: http://example.com/files/foo
      * entry with files: http://example.com/files/foo/bar.txt
                          http://example.com/files/foo/moo.hdf5"""
    if entry and entry.get('files'):
        return [f for f in entry['files'] if url in f['url']]
    return []


def get_relative_filenames(url, entry):
    base_url = os.path.dirname(url)
    return [f['url'].replace(base_url, '').lstrip('/')
            for f in get_matching_files(url, entry)]


def get_paths(entry):
    return [urllib.parse.urlparse(f.get('url')).path
            for f in entry.get('files', [])]


def is_top_level(entry, subject):
    sub_path = urllib.parse.urlparse(subject).path
    return all([f_path.startswith(sub_path) for f_path in get_paths(entry)])



import sys
import click
import logging
import globus_sdk

from pilot import commands, exc
from pilot.version import __version__
from pilot.commands.auth import auth_commands
from pilot.commands.search import search_commands, delete
from pilot.commands.transfer import transfer_commands, status_commands, analyze
from pilot.commands.project import project, index

log = logging.getLogger(__name__)

INVOKABLE_WITHOUT_LOGIN = ['login', 'logout', 'version']


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    pc = commands.get_pilot_client()

    if not pc.config.is_migrated():
        click.secho('Old config detected, upgrading... ', fg='yellow',
                    nl=False)
        try:
            pc.config.migrate()
            click.secho('Success!', fg='green')
        except Exception:
            click.secho(f'Failed! Try removing '
                        f'{pc.config_file} and logging in '
                        f'again.', fg='red')
    if pc.is_logged_in():
        if pc.context.is_cache_stale():
            log.debug('Cache is stale! Updating...')
            try:
                if any(pc.context.update_with_diff(dry_run=True).values()):
                    click.secho('Projects have updated. Use '
                                '"pilot project update"'
                                ' to get the newest changes.', fg='yellow')
                log.debug('Update was successful.')
            except globus_sdk.exc.SearchAPIError as sapie:
                if not sapie.code == 'NotFound.Generic':
                    log.error('This error is unexpected!')
                    log.exception(sapie)
                click.secho(
                    'No manifest exists on this index. You can add '
                    'one by running `pilot context push`', fg='red')
        pcommands = ['delete', 'describe', 'download', 'list', 'mkdir',
                     'upload']
        if not pc.project.is_set() and ctx.invoked_subcommand in pcommands:
            click.secho('No project set, use "pilot project" to list projects '
                        'and "pilot project set <myproject>" '
                        'to set your current project.', fg='yellow')
            sys.exit(exc.ExitCodes.INVALID_CLIENT_CONFIGURATION)
    else:
        if (ctx.invoked_subcommand and
                ctx.invoked_subcommand not in INVOKABLE_WITHOUT_LOGIN):
            click.echo('You are not logged in.')
            sys.exit(exc.ExitCodes.NOT_LOGGED_IN)

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@click.command(help='Show version and exit')
def version():
    click.echo(__version__)


cli.add_command(auth_commands.login)
cli.add_command(auth_commands.logout)
cli.add_command(auth_commands.profile_command)

cli.add_command(project.project_command)
cli.add_command(index.index_command)

cli.add_command(search_commands.list_command)
cli.add_command(search_commands.describe)
cli.add_command(delete.delete_command)

cli.add_command(transfer_commands.upload)
cli.add_command(analyze.analyze)
cli.add_command(transfer_commands.download)
cli.add_command(transfer_commands.mkdir)
cli.add_command(transfer_commands.register)
cli.add_command(status_commands.status_command)

cli.add_command(version)



import os
import pytest
import globus_sdk
from unittest.mock import Mock, call, mock_open, patch
from pilot import globus_clients, exc
from tests.unit.mocks import (MOCK_PROFILE, MOCK_TOKEN_SET,
                              CLIENT_FILE_BASE_DIR)
from fair_research_login.exc import LoadError

TINY_DATAFRAME = os.path.join(CLIENT_FILE_BASE_DIR, 'tiny_dataframe.tsv')


def test_login(monkeypatch, mock_cli_basic, mock_config):
    assert mock_config.load()['profile'] == {}
    auth_cli = Mock()
    monkeypatch.setattr(mock_cli_basic, 'get_auth_client',
                        Mock(return_value=auth_cli))
    mp = MOCK_PROFILE.copy()
    mp['name'] = 'Real Genius'
    auth_cli.oauth2_userinfo.return_value.data = mp
    mock_cli_basic.login()
    assert auth_cli.oauth2_userinfo.called
    assert mock_cli_basic.profile.name == mp['name']


def test_logout(monkeypatch, mock_cli_basic, mock_config):
    monkeypatch.setattr(mock_cli_basic, 'revoke_token_set', Mock())
    cfg = mock_config.load()
    cfg['tokens'] = MOCK_TOKEN_SET
    mock_cli_basic.logout()
    assert mock_config.load()['tokens'] == {}


def test_is_logged_in(monkeypatch, mock_cli_basic):
    load_tokens = Mock()
    monkeypatch.setattr(mock_cli_basic, 'load_tokens', load_tokens)

    load_tokens.return_value = {'a bunch ': 'of tokens'}
    assert mock_cli_basic.is_logged_in() is True

    load_tokens.side_effect = LoadError
    assert mock_cli_basic.is_logged_in() is False


def test_get_clients(mock_cli_basic):
    assert isinstance(mock_cli_basic.get_auth_client(), globus_sdk.AuthClient)
    assert isinstance(mock_cli_basic.get_transfer_client(),
                      globus_sdk.TransferClient)
    assert isinstance(mock_cli_basic.get_search_client(),
                      globus_sdk.SearchClient)


def test_ls(monkeypatch, mock_cli_basic):
    transfer_cli = Mock()
    transfer_cli.operation_ls.return_value = {
        'DATA': [{'type': 'file', 'name': 'one'},
                 {'type': 'dir', 'name': 'two'}]
    }
    monkeypatch.setattr(mock_cli_basic, 'get_transfer_client',
                        Mock(return_value=transfer_cli))
    result = mock_cli_basic.ls('foo')

    assert transfer_cli.operation_ls.called
    transfer_cli.operation_ls.assert_called_with('foo-project-endpoint',
                                                 path='/foo_folder/foo')
    # ensure only the file was returned
    assert len(result) == 2


def test_upload_http(monkeypatch, mixed_tsv, mock_cli_basic):
    put = Mock()
    monkeypatch.setattr(globus_clients.HTTPFileClient, 'put', put)
    mock_cli_basic.upload_http(TINY_DATAFRAME, 'destination')
    assert put.called
    assert put.call_args == call('/foo_folder/destination/tiny_dataframe.tsv',
                                 filename=TINY_DATAFRAME)


def test_download_http(monkeypatch, mixed_tsv, mock_cli_basic, mock_projects):
    response = Mock()
    response.iter_content = ['hello', 'world']
    get = Mock(return_value=response)
    monkeypatch.setattr(globus_clients.HTTPFileClient, 'get', get)
    monkeypatch.setattr(os, 'mkdir', Mock())
    m_open = mock_open()
    m_open.return_value.write.return_value = 1
    with patch("builtins.open", m_open):
        mock_cli_basic.download('a.tsv')
    assert get.called
    assert get.call_args == call('/foo_folder/a.tsv', range=None)


def test_ingest(monkeypatch, mock_cli_basic, mock_sdk_response):
    search_cli = Mock()
    search_cli.ingest.return_value = {'task_id': 'foo'}
    search_cli.get_task = Mock(return_value={'state': 'SUCCESS'})
    monkeypatch.setattr(mock_cli_basic, 'get_search_client',
                        Mock(return_value=search_cli))
    meta = mock_cli_basic.gather_metadata(TINY_DATAFRAME, '/')
    r = mock_cli_basic.ingest('tiny_dataframe.tsv', meta)
    assert r is True

    mock_sdk_response.data = {'state': 'FAILURE', 'message': 'it failed!'}
    search_cli.get_task.return_value = mock_sdk_response
    with pytest.raises(exc.PilotClientException):
        mock_cli_basic.ingest('tiny_dataframe.tsv', meta)


def test_get_search_entry(monkeypatch, mock_cli_basic):
    search_cli = Mock()
    search_cli.get_subject.return_value = {'content': ['myresult']}
    mock_cli_basic.get_search_client = Mock(return_value=search_cli)

    assert mock_cli_basic.get_search_entry('foo') == 'myresult'
    search_cli.get_subject.assert_called_with(
        'foo-search-index',
        'globus://foo-project-endpoint/foo_folder/foo',
        result_format_version='2017-09-01',
    )

    class MockException(Exception):
        code = 'random_exception.generic'

    monkeypatch.setattr(globus_sdk.exc, 'SearchAPIError', MockException)
    search_cli.get_subject.side_effect = globus_sdk.exc.SearchAPIError
    assert mock_cli_basic.get_search_entry('foo') is None


def test_list_entries(mock_multi_file_result, mock_cli_basic):
    results = mock_multi_file_result.copy()
    mfr = mock_multi_file_result['gmeta'][0]
    ent2 = mfr.copy()
    ent2['subject'] = 'globus://foo-project-endpoint/foo_folder/foo/bar/moo'
    ent3 = mfr.copy()
    ent3['subject'] = 'globus://foo-project-endpoint/foo_folder/foo/bar/baz'
    results['gmeta'] = [mfr, ent2, ent3]
    ps_response = Mock()
    ps_response.data = results
    search_cli = Mock()
    mock_cli_basic.get_search_client = Mock(return_value=search_cli)
    search_cli.post_search.return_value = ps_response

    assert len(mock_cli_basic.list_entries('')) == 3
    assert len(mock_cli_basic.list_entries('foo/bar')) == 2
    assert len(mock_cli_basic.list_entries('foo/bar/moo')) == 1
    assert len(mock_cli_basic.list_entries('foo/bar/baz')) == 1
    assert len(mock_cli_basic.list_entries('foo/foo/foo')) == 0


def test_get_search_entry_dir(monkeypatch, mock_cli_basic,
                              mock_multi_file_result):
    search_cli = Mock()
    mock_cli_basic.get_search_client = Mock(return_value=search_cli)

    class MockException(Exception):
        code = 'NotFound.Generic'

    monkeypatch.setattr(globus_sdk.exc, 'SearchAPIError', MockException)
    search_cli.get_subject.side_effect = globus_sdk.exc.SearchAPIError
    ps_response = Mock()
    ps_response.data = mock_multi_file_result
    search_cli.post_search.return_value = ps_response
    entry = mock_cli_basic.get_search_entry('multi_file/text_metadata.txt')
    assert entry == mock_multi_file_result['gmeta'][0]['content'][0]


def test_delete_entry_sub(monkeypatch, mock_cli_basic, mock_multi_file_result):
    search_cli = Mock()
    mock_cli_basic.ingest_entry = Mock()
    monkeypatch.setattr(mock_cli_basic, 'get_search_client',
                        Mock(return_value=search_cli))
    ment = Mock(return_value=mock_multi_file_result['gmeta'][0])
    mock_cli_basic.get_full_search_entry = ment
    mock_cli_basic.delete_entry('/multi_file', entry_id='my_id')
    assert not mock_cli_basic.ingest_entry.called
    assert search_cli.delete_entry.call_args == call(
        'foo-search-index',
        'globus://foo-project-endpoint/foo_folder/multi_file',
        entry_id='my_id'
    )
    mock_cli_basic.delete_entry('/multi_file', 'my_custom_id',
                                full_subject=True)
    assert search_cli.delete_subject.called


def test_delete_file_in_mfe(mock_cli_basic, mock_search_client,
                            mock_multi_file_result):
    mock_cli_basic.ingest = Mock()
    mock_cli_basic.get_full_search_entry = Mock(
        return_value=mock_multi_file_result['gmeta'][0])
    mock_cli_basic.delete_entry('/multi_file/text_metadata.txt')
    assert mock_cli_basic.ingest.called


def test_delete_entry_no_result(monkeypatch, mock_cli_basic):
    search_cli = Mock()
    monkeypatch.setattr(mock_cli_basic, 'get_search_client',
                        Mock(return_value=search_cli))
    mock_cli_basic.get_full_search_entry = Mock(return_value=None)
    with pytest.raises(exc.RecordDoesNotExist):
        mock_cli_basic.delete_entry('foo', 'bar')



import os
import json
from unittest.mock import Mock
from click.testing import CliRunner
import globus_sdk
from pilot.commands.transfer.transfer_commands import upload
from pilot import transfer_log
from pilot.search import scrape_metadata
from pilot import analysis
from pilot.exc import ExitCodes, AnalysisException
from tests.unit.mocks import COMMANDS_FILE_BASE_DIR

EMPTY_TEST_FILE = os.path.join(COMMANDS_FILE_BASE_DIR,
                               'test_file_zero_length.txt')
EMTPY_TEST_FILE_META = os.path.join(COMMANDS_FILE_BASE_DIR,
                                    'empty_file_metadata.json')
SMALL_TEST_FILE = os.path.join(COMMANDS_FILE_BASE_DIR, 'test_file_small.txt')
CUSTOM_METADATA = os.path.join(COMMANDS_FILE_BASE_DIR, 'custom_metadata.json')
INVALID_METADATA = os.path.join(COMMANDS_FILE_BASE_DIR,
                                'invalid_metadata.json')


def test_upload(mock_cli, monkeypatch):
    monkeypatch.setattr(transfer_log.TransferLog, 'add_log', Mock())
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder'])
    assert result.exit_code == ExitCodes.SUCCESS


def test_upload_without_destination(mock_cli):
    mock_cli.ls.return_value = {'foo': {'type': 'dir'},
                                'bar': {'type': 'file'}}
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE])
    assert result.exit_code == ExitCodes.NO_DESTINATION_PROVIDED
    assert 'No Destination Provided' in result.output
    assert 'foo' in result.output
    assert 'bar' not in result.output


def test_upload_to_nonexistant_dir(mock_cli, mock_transfer_error):
    mock_transfer_error.code = 'ClientError.NotFound'
    mock_cli.ls = Mock(side_effect=globus_sdk.exc.TransferAPIError)
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder'])
    assert result.exit_code == ExitCodes.DIRECTORY_DOES_NOT_EXIST


def test_upload_unexpected_ls_error(mock_cli, mock_transfer_error):
    mock_transfer_error.code = 'UnexpectedError'
    mock_cli.ls = Mock(side_effect=globus_sdk.exc.TransferAPIError)
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder'])
    assert result.exit_code == ExitCodes.GLOBUS_TRANSFER_ERROR


def test_upload_with_custom_metadata(mock_cli):
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder',
                                         '-j', CUSTOM_METADATA, '--verbose',
                                         '--dry-run'])
    assert result.exit_code == 0
    assert 'custom_metadata_key' in result.output
    assert 'custom_metadata_value' in result.output


def test_upload_analyze_error(mock_cli, monkeypatch):
    mock_exc = Mock(side_effect=AnalysisException('fail!', None))
    monkeypatch.setattr(analysis, 'analyze_dataframe', mock_exc)
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder',
                                         '--no-gcp'])
    assert result.exit_code == 0
    assert 'Error analyzing' in result.output
    assert '(Use --verbose to see full error)' in result.output


def test_upload_validation_error(mock_cli):
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder',
                                         '--no-gcp', '-j', INVALID_METADATA])
    print(result.output)
    assert result.exit_code == ExitCodes.INVALID_METADATA


def test_no_update_needed(mock_cli, mock_search_results):
    base_name = os.path.basename(EMPTY_TEST_FILE)
    url = mock_cli.get_globus_http_url(base_name)
    sub = mock_cli.get_subject_url(base_name)
    meta = scrape_metadata(EMPTY_TEST_FILE, url, mock_cli.profile,
                           'foo-project')
    mock_search_results['gmeta'][0]['content'][0] = meta
    mock_search_results['gmeta'][0]['subject'] = sub
    mock_cli.list_entries = Mock(return_value=mock_search_results['gmeta'])
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, '/',
                                         '--no-gcp', '-u'])
    assert result.exit_code == 0
    assert 'Files and search entry are an exact match.' in result.output


def test_upload_record_exists(mock_cli, mock_search_results):
    base_name = os.path.basename(EMPTY_TEST_FILE)
    url = mock_cli.get_globus_http_url(base_name)
    sub = mock_cli.get_subject_url(base_name)
    meta = scrape_metadata(EMPTY_TEST_FILE, url, mock_cli.profile,
                           'foo-project')
    mock_search_results['gmeta'][0]['content'][0] = meta
    mock_search_results['gmeta'][0]['subject'] = sub
    mock_cli.list_entries = Mock(return_value=mock_search_results['gmeta'])

    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, '/', '--no-gcp'])
    assert result.exit_code == ExitCodes.RECORD_EXISTS


def test_upload_dry_run(mock_cli):
    result = CliRunner().invoke(upload, [SMALL_TEST_FILE, 'my_folder',
                                         '--verbose', '--dry-run'])
    assert result.exit_code == 0
    assert 'text/plain' in result.output


def test_dataframe_up_to_date(mock_cli, mock_transfer_log,
                              mock_search_results):
    with open(EMTPY_TEST_FILE_META) as f:
        meta = json.load(f)
    base_name = os.path.basename(EMPTY_TEST_FILE)
    sub = mock_cli.get_subject_url(base_name)
    mock_search_results['gmeta'][0]['content'][0] = meta
    mock_search_results['gmeta'][0]['subject'] = sub
    mock_cli.list_entries = Mock(return_value=mock_search_results['gmeta'])
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, '/',
                                         '-u', '-j', CUSTOM_METADATA])
    assert result.exit_code == 0
    assert not globus_sdk.TransferData.called


def test_upload_local_endpoint_not_set(mock_cli, mock_profile):

    mock_cli.profile.save_option('local_endpoint', None, section='profile')
    result = CliRunner().invoke(upload, [SMALL_TEST_FILE, 'my_folder'
                                         '--gcp'])
    assert result.exit_code == ExitCodes.NO_LOCAL_ENDPOINT_SET


def test_upload_gcp_log(mock_cli, mock_transfer_log):
    mock_cli.get_transfer_client().submit_transfer.return_value = {}
    result = CliRunner().invoke(upload, [EMPTY_TEST_FILE, 'my_folder'])
    assert result.exit_code == 0
    assert mock_transfer_log.called



from pilot.client import PilotClient
from pilot import logging_cfg

import logging

log = logging.getLogger(__name__)


def get_pilot_client():
    logging_cfg.setup_logging(level='CRITICAL')
    return PilotClient()



import os
import urllib
import logging
import datetime

log = logging.getLogger(__name__)

FIELD_METADATA_TITLES = [
    ('name', 'Column Name'),
    ('type', 'Type'), ('count', 'Count'),
    ('frequency', 'Freq'), ('top', 'Top'),
    ('unique', 'Unique'),
    ('min', 'Min'),
    ('max', 'Max'),
    ('mean', 'Mean'),
    ('std', 'Std'),
    ('25', '25-PCTL'),
    ('50', '50-PCTL'),
    ('75', '75-PCTL')
 ]


def get_formatted_fields(entry, columns, formatting='{:21.20}{}', limit=10):
    output = []
    raw_data = dict(parse_result(entry, columns))
    tdata = zip(get_titles(columns),
                [raw_data[name] for name in columns])
    for title, data in tdata:
        if isinstance(data, list):
            if len(data) > limit:
                data.insert(0, 'List truncated due to number of items. '
                               'Showing 10/{}.'.format(len(data)))
                data = data[:limit]
            output += [formatting.format(title, line) for line in data[:1]]
            output += [formatting.format('', line) for line in data[1:]]
        else:
            output.append(formatting.format(title, data))
    return output


def get_formatted_field_metadata(field_metadata):
    fmt = ('{:21.20}'
           '{:8.7}{:7.6}{:5.4}{:12.11}{:7.6}'
           '{:7.6}{:8.7}{:8.7}{:8.7}')
    _, titles = zip(*FIELD_METADATA_TITLES)
    formatted_titles = [fmt.format(*titles)]

    data = []
    for row in get_field_metadata(field_metadata):
        _, fm_data = zip(*row)
        data.append(fmt.format(*[str(i) for i in fm_data]))
    if data:
        return formatted_titles + data
    return []


def get_titles(list_of_names):
    title_map = dict([(name, title) for name, title, _ in GENERAL_PARSE_FUNCS])
    return [title_map.get(name) for name in list_of_names]


def parse_result(result, fields=None):
    processed_results = []
    fields = fields or GENERAL_PARSE_FUNCS
    funcs_subset = [func for func in GENERAL_PARSE_FUNCS if func[0] in fields]
    for name, _, processor in funcs_subset:
        try:
            data = processor(result)
        except Exception:
            # log.exception(e)
            data = ''
        processed_results.append((name, data))
    return processed_results


def get_field_metadata(field_metadata):
    ret = []
    metadata_names = [n for n, _ in FIELD_METADATA_TITLES]
    for field_metadata_item in field_metadata.get('field_definitions', []):
        ret.append([(name, field_metadata_item.get(name, ''))
                    for name in metadata_names])
    return ret


def get_size(result):
    size = sum([f.get('length', 0) for f in result.get('files')])
    # 2**10 = 1024
    power = 2**10
    n = 0
    Dic_powerN = {0: '', 1: 'K', 2: 'M', 3: 'G', 4: 'T'}
    while size > power:
        size /= power
        n += 1

    return '{} {}'.format(int(size), Dic_powerN[n])


def get_dates(result):
    dates = result['dc']['dates']
    fdates = []
    for date in dates:
        dt = datetime.datetime.strptime(date['date'], '%Y-%m-%dT%H:%M:%S.%fZ')
        fdates.append('{}: {: %A, %b %d, %Y}'.format(date['dateType'], dt))
    return fdates


def get_paths(result):
    return [urllib.parse.urlparse(f.get('url')).path
            for f in result.get('files')]


def get_common_path(result):
    paths = get_paths(result)
    common_path = paths[0]
    while common_path and not all([common_path in p for p in paths]):
        common_path = os.path.dirname(common_path)
    return common_path


def get_relative_paths(result):
    return [path.replace(get_common_path(result), '').lstrip('/')
            for path in get_paths(result)]


def get_formats(result):
    formats = [f['mime_type'] for f in result.get('files', {})
               if f.get('mime_type')]
    return ['{} ({})'.format(f, formats.count(f)) for f in set(formats)]


def get_files(result):
    listings = []
    common_path = get_common_path(result)
    for f in result.get('files', []):
        path = urllib.parse.urlparse(f.get('url')).path
        path = path.replace(common_path, '').lstrip('/')
        listings.append('{} ({})'.format(path, f.get('mime_type')))
    return sorted(listings)


GENERAL_PARSE_FUNCS = [
    ('title', 'Title', lambda r: r['dc']['titles'][0]['title']),
    ('authors', 'Authors',
     lambda r: [c['creatorName'] for c in r['dc']['creators']]),
    ('publisher', 'Publisher', lambda r: r['dc']['publisher']),
    ('subjects', 'Subjects',
     lambda r: [s['subject'] for s in r['dc']['subjects']]),
    ('dates', 'Dates', get_dates),
    ('data', 'Data',
     lambda r: r['project_metadata']['data_type']),
    ('dataframe', 'Dataframe',
     lambda r: r['project_metadata']['dataframe_type']),
    ('rows', 'Rows',
     lambda r: r['files'][0]['field_metadata']['numrows']),
    ('columns', 'Columns',
     lambda r: r['files'][0]['field_metadata']['numcols']),
    ('formats', 'Formats', get_formats),
    ('version', 'Version', lambda r: r['dc']['version']),
    ('size', 'Size', get_size),
    ('combined_size', 'Combined Size', get_size),
    ('files', 'Files', get_files),
    ('description', 'Description',
     lambda r: r['dc']['descriptions'][0]['description']),
]
